from petram.mfem_config import use_parallel
from .solver_model import LinearSolverModel, LinearSolver
from petram.helper.matrix_file import write_matrix, write_vector, write_coo_matrix
import os
import numpy as np
import scipy
import weakref

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('MUMPSModel')


def convert2float(txt):
    try:
        return float(txt)
    except BaseException:
        assert False, "can not convert to float. Input text is " + txt


def convert2int(txt):
    try:
        return int(txt)
    except BaseException:
        assert False, "can not convert to float. Input text is " + txt


class MUMPSBase(LinearSolverModel):
    has_2nd_panel = False
    accept_complex = True
    is_iterative = False

    def __init__(self):
        self.s = None
        LinearSolverModel.__init__(self)

    def init_solver(self):
        pass

    def panel1_param(self):
        return [["log_level(0-2)", self.log_level, 400, {}],
                ["ordering", self.ordering, 4, {"readonly": True,
                                                "choices": ["auto", "AMF", "PORD", "QAMD",
                                                            "Metis", "Scotch",
                                                            "ParMetis", "PT-Scotch"]}],
                ["out-of-core", self.out_of_core, 3, {"text": ""}],
                ["error analysis", self.error_ana, 4, {"readonly": True,
                                                       "choices": ["none", "full stat.", "main stat."]}],
                ["write matrix", self.write_mat, 3, {"text": ""}],
                ["write factor", self.write_fac, 3, {"text": ""}],
                ["restore factor", self.restore_fac, 3, {"text": ""}],
                ["factor path/prefix", self.factor_path, 0, {}],
                ["use BLR", self.use_blr, 3, {"text": ""}],
                ["BLR drop parameter", self.blr_drop, 300, {}],
                ["WS Inc. (ICNTL14)", self.icntl14, 0, {}],
                ["WS Size (ICNTL23)", self.icntl23, 0, {}],
                ["numerical pivot thr. (CNTL1)", self.cntl1, 0, {}],
                ["static pivot thr. (CNTL4)", self.cntl4, 0, {}],
                ["Itr. refinement (ICNTL10)", self.icntl10, 0, {}],
                ["refinement stop Cond. (CNTL2)", self.cntl2, 0, {}],
                ["permutation/scaling Opt.(ICNTL6)", self.icntl6, 0, {}],
                ["scaling strategy (ICNTL8)", self.icntl8, 0, {}],
                ["write inverse", self.write_inv, 3, {"text": ""}],
                ["use float32", self.use_single_precision, 3, {"text": ""}],
                ["use dist, RHS (only for MLsolver)",
                 self.use_dist_rhs, 3, {"text": ""}],
                ["use dist, SOL (dev.)", self.use_dist_sol, 3, {"text": ""}], ]

    def get_panel1_value(self):
        return (int(self.log_level), self.ordering, self.out_of_core,
                self.error_ana, self.write_mat, self.write_fac,
                self.restore_fac, self.factor_path,
                self.use_blr, self.blr_drop, str(self.icntl14),
                str(self.icntl23),
                self.cntl1, self.cntl4, self.icntl10, self.cntl2,
                self.icntl6, self.icntl8, self.write_inv,
                self.use_single_precision, self.use_dist_rhs, self.use_dist_sol)

    def import_panel1_value(self, v):
        self.log_level = int(v[0])
        self.ordering = str(v[1])
        self.out_of_core = v[2]
        self.error_ana = v[3]
        self.write_mat = v[4]
        self.write_fac = v[5]
        self.restore_fac = v[6]
        self.factor_path = v[7]
        self.use_blr = v[8]
        self.blr_drop = v[9]
        self.icntl14 = v[10]
        self.icntl23 = v[11]
        self.cntl1 = v[12]
        self.cntl4 = v[13]
        self.icntl10 = v[14]
        self.cntl2 = v[15]
        self.icntl6 = v[16]
        self.icntl8 = v[17]
        self.write_inv = v[18]
        self.use_single_precision = v[19]
        self.use_dist_rhs = v[20]
        self.use_dist_sol = v[21]

    def attribute_set(self, v):
        v = super(MUMPSBase, self).attribute_set(v)

        v['log_level'] = 0
        '''
        1 : Only error messages printed.
        2 : Errors, warnings, and main statistics printed.
        3 : Errors and warnings and terse diagnostics
            (only first ten entries of arrays) printed.
        4 : Errors, warnings and information on input, output parameters printed
        '''
        v['out_of_core'] = False
        v['write_mat'] = False
        v['write_fac'] = False
        v['restore_fac'] = False
        v['factor_path'] = "~/mumps_data/factored_matrix*"
        v['central_mat'] = False
        v['ordering'] = 'auto'
        v['error_ana'] = 'none'
        v['use_blr'] = False
        v['blr_drop'] = 0.0
        v['icntl14'] = '20'
        v['icntl23'] = '0'
        v['cntl1'] = 'default'
        v['cntl4'] = 'default'
        v['icntl10'] = 'default'
        v['cntl2'] = 'default'
        v['icntl6'] = 'default'
        v['icntl8'] = 'default'
        v['use_single_precision'] = False
        v['write_inv'] = False
        v['use_dist_rhs'] = False
        v['use_dist_sol'] = True

        # make sure that old data type (data was stored as int) is converted to
        # string
        if hasattr(self, "icntl14"):
            self.icntl14 = str(self.icntl14)
        if hasattr(self, "icntl23"):
            self.icntl23 = str(self.icntl23)

        # this flag needs to be set, so that destcuctor works when
        # model tree is loaded from pickled file
        v['s'] = None
        return v

    def allocate_solver(self, is_complex=False, engine=None):
        # engine not used
        ls_type = self.get_solve_root().get_linearsystem_type_from_modeltree()

        if ls_type.startswith('blk'):
            solver = MUMPSBlockSolver(self, engine)
        else:
            solver = MUMPSSolver(self, engine)
            solver.AllocSolver(is_complex, self.use_single_precision)
        return solver

    def solve(self, engine, A, b):
        solver = self.allocate_solver((A.dtype == 'complex'), engine)
        solver.SetOperator(A, dist=engine.is_matrix_distributed)
        solall = solver.Mult(b, case_base=engine.case_base)
        return solall

    def real_to_complex(self, solall, M=None):
        try:
            from mpi4py import MPI
        except BaseException:
            from petram.helper.dummy_mpi import MPI
        myid = MPI.COMM_WORLD.rank
        nproc = MPI.COMM_WORLD.size

        if myid == 0:
            s = solall.shape[0]
            solall = solall[:s // 2, :] + 1j * solall[s // 2:, :]
            return solall

    def __del__(self):
        if self.s is not None:
            self.s.finish()
        self.s = None


class MUMPS(MUMPSBase):
    def does_linearsolver_choose_linearsystem_type(self):
        from petram.solver.solver_model import Solver

        # for the sake of backword compabitibiliy. it is true only when
        # it is the top level linearsolver

#        if isinstance(self.parent, Solver):
        from petram.solver.solver_controls import SolveControl

        if self.parent.parent is self.get_solve_root():
            return True
        elif isinstance(self.parent.parent, SolveControl):
            if self.parent.parent.parent is self.get_solve_root():
                return True
            else:
                return False
        else:
            return False

    def supported_linear_system_type(self):
        return 'ANY'

    def linear_system_type(self, assemble_real, phys_real):
        if phys_real:
            return 'coo'
        if assemble_real:
            return 'coo_real'
        return 'coo'


class MUMPSMFEMSolverModel(MUMPSBase):
    '''
    This one is to use MUMPS in iterative solver
    It creates MUMPSBlockPreconditioner
    '''
    @classmethod
    def fancy_menu_name(self):
        return 'MUMPS'

    @classmethod
    def fancy_tree_name(self):
        return 'MUMPS'

    def does_linearsolver_choose_linearsystem_type(self):
        return False

    def supported_linear_system_type(self):
        return ["blk_interleave",
                "blk_merged_s",
                "blk_merged", ]

    def prepare_solver(self, opr, engine):
        solver = MUMPSBlockPreconditioner(opr,
                                          gui=self,
                                          engine=engine,)
        solver.SetOperator(opr)
        return solver


class MUMPSPreconditionerModel(MUMPSBase):
    '''
    This one is to use MUMPS in iterative solver
    It creates MUMPSPreconditioner
    '''
    @classmethod
    def fancy_menu_name(self):
        return 'MUMPS'

    @classmethod
    def fancy_tree_name(self):
        return 'MUMPS'

    def does_linearsolver_choose_linearsystem_type(self):
        return False

    def supported_linear_system_type(self):
        return ["blk_interleave",
                "blk_merged_s",
                "blk_merged", ]

    def prepare_solver(self, opr, engine):
        prc = MUMPSPreconditioner(opr,
                                  gui=self,
                                  engine=engine)
        return prc


class MUMPSSolver(LinearSolver):
    is_iterative = False

    def __init__(self, *args, **kwargs):
        super(MUMPSSolver, self).__init__(*args, **kwargs)
        self.silent = False
        self.keep_sol_distributed = False

    @staticmethod
    def split_dir_prefix(txt):
        txt = os.path.abspath(os.path.expanduser(txt))
        txt = txt.strip().split('*')[0]
        path = os.path.dirname(txt)
        prefix = os.path.basename(txt)
        try:
            from mpi4py import MPI
            myid = MPI.COMM_WORLD.rank
        except BaseException:
            myid = 0
            from petram.helper.dummy_mpi import MPI

        if myid == 0:
            if not os.path.exists(path):
                dprint1("Creating path for MUMPS data", path)
                os.makedirs(path)
        return prefix, path

    def set_silent(self, silent):
        self.silent = silent

    def set_ordering_flag(self, s):
        from petram.mfem_config import use_parallel
        gui = self.gui
        if gui.ordering == 'auto':
            pass
        elif gui.ordering == 'QAMD':
            # if use_parallel:
            #    dprint1(
            #        "!!! QAMD ordering is selected. But solver is not in serial mode. Ignored")
            # else:
            s.set_icntl(28, 1)
            s.set_icntl(7, 6)
        elif gui.ordering == 'AMF':
            # if use_parallel:
            #    dprint1(
            #        "!!! AMF ordering is selected. But solver is not in serial mode. Ignored")
            # else:
            s.set_icntl(28, 1)
            s.set_icntl(7, 2)
        elif gui.ordering == 'PORD':
            # if use_parallel:
            #    dprint1(
            #        "!!! PORD ordering is selected. But solver is not in serial mode. Ignored")
            # else:
            s.set_icntl(28, 1)
            s.set_icntl(7, 4)
        elif gui.ordering == 'Metis':
            # if use_parallel:
            #    dprint1(
            #        "!!! Metis ordering is selected. But solver is not in serial mode. Ignored")
            # else:
            s.set_icntl(28, 1)
            s.set_icntl(7, 5)
        elif gui.ordering == 'Scotch':
            # if use_parallel:
            #    dprint1(
            #        "!!! Scotch ordering is selected. But solver is not in serial mode. Ignored")
            # else:
            s.set_icntl(28, 1)
            s.set_icntl(7, 3)
        elif gui.ordering == 'ParMetis':
            if use_parallel:
                s.set_icntl(28, 2)
                s.set_icntl(29, 2)
            else:
                dprint1(
                    "!!! ParMetis ordering is selected. But solver is not in parallel mode. Ignored")
        elif gui.ordering == 'PT-Scotch':
            if use_parallel:
                s.set_icntl(28, 2)
                s.set_icntl(29, 1)
            else:
                dprint1(
                    "!!! PT-Scotch ordering is selected. But solver is not in parallel mode. Ignored")
        else:
            pass
        #s.set_icntl(28,  2)

    def set_error_analysis(self, s):
        from petram.mfem_config import use_parallel
        gui = self.gui
        if gui.error_ana == 'none':
            s.set_icntl(11, 0)
        elif gui.error_ana == 'main stat.':
            s.set_icntl(11, 2)
        elif gui.error_ana == 'full stat.':
            s.set_icntl(11, 1)
        else:
            pass

    def AllocSolver(self, is_complex, use_single_precision):
        if use_parallel:
            from mpi4py import MPI
            myid = MPI.COMM_WORLD.rank
            args = (1, 0, MPI.COMM_WORLD)
        else:
            myid = 0
            args = (1, 0)

        if is_complex:
            if use_single_precision:
                from petram.ext.mumps.mumps_solve import c_array as data_array
                from petram.ext.mumps.mumps_solve import CMUMPS
                s = CMUMPS(*args)
            else:
                from petram.ext.mumps.mumps_solve import z_array as data_array
                from petram.ext.mumps.mumps_solve import ZMUMPS
                s = ZMUMPS(*args)
        else:
            if use_single_precision:
                from petram.ext.mumps.mumps_solve import s_array as data_array
                from petram.ext.mumps.mumps_solve import SMUMPS
                s = SMUMPS(*args)
            else:
                from petram.ext.mumps.mumps_solve import d_array as data_array
                from petram.ext.mumps.mumps_solve import DMUMPS
                s = DMUMPS(*args)

        self.s = s
        self.is_complex = is_complex
        self.data_array = data_array

        gui = self.gui
        # No outputs
        if gui.log_level == 0:
            s.set_icntl(1, -1)
            s.set_icntl(2, -1)
            s.set_icntl(3, -1)
            s.set_icntl(4, 0)
        elif gui.log_level == 1:
            pass
        else:
            if myid == 0:
                s.set_icntl(1, 6)
                s.set_icntl(2, 6)
                s.set_icntl(3, 6)
                s.set_icntl(4, 6)

    def _data_type(self):
        if self.gui.use_single_precision:
            if self.is_complex:
                return np.complex64
            else:
                return np.float32
        else:
            if self.is_complex:
                return np.complex128
            else:
                return np.float64

    def _int_type(self):
        import petram.ext.mumps.mumps_solve as mumps_solve
        dtype_int = 'int' + str(mumps_solve.SIZEOF_MUMPS_INT() * 8)
        return dtype_int

    def make_matrix_entries(self, A):
        datatype = self._data_type()
        AA = A.data.astype(datatype, copy=False)
        return AA
        '''
        if self.gui.use_single_precision:
            if self.is_complex:
                AA = A.data.astype(np.complex64, copy=False)
            else:
                AA = A.data.astype(np.float32, copy=False)
        else:
            if self.is_complex:
                AA = A.data.astype(np.complex128, copy=False)
            else:
                AA = A.data.astype(np.float64, copy=False)
        return AA
        '''

    def make_vector_entries(self, B):
        datatype = self._data_type()
        return B.astype(datatype, copy=False)
        '''
        if self.gui.use_single_precision:
            if self.is_complex:
                return B.astype(np.complex64, copy=False)
            else:
                return B.astype(np.float32, copy=False)
        else:
            if self.is_complex:
                return B.astype(np.complex128, copy=False)
            else:
                return B.astype(np.float64, copy=False)
        '''

    def _merge_coo_matrix(self, myid, rank):
        def filename(myid):
            smyid = '{:0>6d}'.format(myid)
            return "matrix." + smyid + ".npz"
        MPI.COMM_WORLD.Barrier()

        if myid != 0:
            return
        f = filename(myid)
        mat = np.load(f, allow_pickle=True)
        os.remove(f)
        mat = mat['A'][()]

        for i in range(1, rank):
            f = filename(i)
            mat2 = np.load(f, allow_pickle=True)
            os.remove(f)
            mat = mat + mat2['A'][()]

        np.savez("matrix", A=mat)

    def set_distributed_sol(self, s, nrhs):
        '''
        setup parameters to receive a distributed solution

        must be called after Factorization
        '''
        from petram.ext.mumps.mumps_solve import i_array

        n_pivots = s.get_info(23)

        sol_loc = np.zeros(n_pivots * nrhs, dtype=self._data_type())
        isol_loc = np.zeros(n_pivots, dtype=self._int_type())

        s.set_sol_loc(self.data_array(sol_loc),
                      n_pivots,
                      i_array(isol_loc))

        s.set_icntl(21, 1)
        return n_pivots, isol_loc, sol_loc

    def read_rhs_distribution(self, s):
        from petram.ext.mumps.mumps_solve import i_array

        lrhs_loc = s.get_info(23)

        irhs_loc = np.empty(lrhs_loc, dtype=self._int_type())

        s.set_irhs_loc(lrhs_loc, i_array(irhs_loc))

        if not self.silent:
            dprint1("job9")
        s.set_job(9)
        s.run()
        return irhs_loc

    def redistributed_array(self, sol, isol_loc, irhs_loc, nrhs):
        '''
        sol[isol_loc] is distributed so that the local sol index
        is irhs_loc
        '''
        from petram.mfem_config import use_parallel
        if not use_parallel:
            assert False, "should not come here in serial"

        from mpi4py import MPI
        from petram.helper.mpi_recipes import alltoall_vector

        nprc = MPI.COMM_WORLD.size
        myid = MPI.COMM_WORLD.rank

        solsend = []
        irhshit = []

        sol = sol.reshape(nrhs, -1)

        for i in range(nprc):
            if i == myid:
                irhs_tmp = MPI.COMM_WORLD.bcast(irhs_loc, root=i)
            else:
                irhs_tmp = MPI.COMM_WORLD.bcast(None, root=i)
            mask = np.in1d(isol_loc, irhs_tmp, assume_unique=True)

            solsend.append(np.array(sol[:, mask].flatten()))
            irhshit.append(np.array(isol_loc[mask]))

        irhs = alltoall_vector(irhshit, int)
        irhs = [] if len(irhs) == 0 else np.hstack(irhs)
        irhs_loc = [] if len(irhs_loc) == 0 else np.hstack(irhs_loc)

        _hoge1, idx1 = np.unique(irhs, return_index=True)
        _hoge2, idx2 = np.unique(irhs_loc, return_inverse=True)

        assert np.all(_hoge1 == _hoge2), "fail here"

        sol2 = alltoall_vector(solsend, sol.dtype)
        sol2 = [s.reshape(nrhs, -1) for s in sol2]
        sol2 = np.hstack(sol2)

        sol2 = np.vstack([sol2[i][idx1][idx2] for i in range(nrhs)])

        return sol2

    def SetOperator(self, A, dist, name=None, ifactor=0):
        try:
            from mpi4py import MPI
        except BaseException:
            from petram.helper.dummy_mpi import MPI

        myid = MPI.COMM_WORLD.rank
        nproc = MPI.COMM_WORLD.size

        from petram.ext.mumps.mumps_solve import i_array
        gui = self.gui
        s = self.s

        if dist:
            dprint1("SetOperator distributed matrix")
            A.eliminate_zeros()
            if gui.write_mat:
                write_coo_matrix('matrix', A)
            if gui.write_inv:
                smyid = '{:0>6d}'.format(myid)
                np.savez("matrix." + smyid, A=A)
                self._merge_coo_matrix(myid, nproc)

            import petram.ext.mumps.mumps_solve as mumps_solve
            dprint1('!!!these two must be consistent')
            dprint1('sizeof(MUMPS_INT) ', mumps_solve.SIZEOF_MUMPS_INT())

            # set matrix format
            s.set_icntl(5, 0)
            s.set_icntl(18, 3)

            dprint1("NNZ local: ", A.nnz)
            nnz_array = np.array(MPI.COMM_WORLD.allgather(A.nnz))
            if myid == 0:
                s.set_n(A.shape[1])
            dprint1("NNZ total: ", np.sum(nnz_array))
            dprint1("NNZ all: ", nnz_array)

            dtype_int = 'int' + str(mumps_solve.SIZEOF_MUMPS_INT() * 8)
            row = A.row
            col = A.col
            row = row.astype(dtype_int) + 1
            col = col.astype(dtype_int) + 1
            AA = self.make_matrix_entries(A)

            if len(col) > 0:
                dprint1('index data size ', type(col[0]))
                dprint1('matrix data type ', type(AA[0]))

            s.set_nz_loc(len(A.data))
            s.set_irn_loc(i_array(row))
            s.set_jcn_loc(i_array(col))
            s.set_a_loc(self.data_array(AA))

            self.dataset = (A.data, row, col)

            self.irhs_loc = np.unique(row)
            self.N_global = np.sum(
                MPI.COMM_WORLD.allgather(len(self.irhs_loc)))

        else:
            A = A.tocoo(False)  # .astype('complex')
            import petram.ext.mumps.mumps_solve as mumps_solve
            dprint1('!!!these two must be consistent')
            dprint1('sizeof(MUMPS_INT) ', mumps_solve.SIZEOF_MUMPS_INT())

            dtype_int = 'int' + str(mumps_solve.SIZEOF_MUMPS_INT() * 8)

            if gui.write_mat:
                # tocsr().tocoo() forces the output is row sorted.
                write_coo_matrix('matrix', A.tocsr().tocoo())
            # No outputs
            if myid == 0:
                row = A.row
                col = A.col
                row = row.astype(dtype_int) + 1
                col = col.astype(dtype_int) + 1
                AA = self.make_matrix_entries(A)

                if len(col) > 0:
                    dprint1('index data size ', type(col[0]))
                    dprint1('matrix data type ', type(AA[0]))

                s.set_n(A.shape[0])
                s.set_nz(len(A.data))
                s.set_irn(i_array(row))
                s.set_jcn(i_array(col))
                s.set_a(self.data_array(AA))
                self.dataset = (A.data, row, col)
                self.irhs_loc = np.unique(row)
                self.N_global = len(self.irhs_loc)
            else:
                self.irhs_loc = None
                self.N_global = None

        if self.skip_solve:
            print("skip solve is on.... returning")
            return

        # blr
        if gui.use_blr:
            s.set_icntl(35, 1)
            s.set_cntl(7, float(gui.blr_drop))

        # out-of-core
        if gui.out_of_core:
            s.set_icntl(22, 1)

        if gui.icntl14.lower() != 'default':
            # percentage increase in the estimated workingspace
            s.set_icntl(14, convert2int(gui.icntl14))

        if gui.icntl23.lower() != 'default':
            # maximum size of the working memory
            s.set_icntl(23, convert2int(gui.icntl23))

        if gui.icntl8.lower() != 'default':
            # the scaling strategy
            s.set_icntl(8, convert2int(gui.icntl8))

        if gui.icntl6.lower() != 'default':
            # permutes the matrix to  azero-freediagonal and/or
            # scale the matrix
            s.set_icntl(6, convert2int(gui.icntl6))

        if gui.cntl1.lower() != 'default':
            # relative threshold for numerical pivoting
            s.set_cntl(1, convert2float(gui.cntl1))

        if gui.cntl4.lower() != 'default':
            # threshold for static pivoting
            s.set_cntl(4, convert2float(gui.cntl4))

        s.set_icntl(24, 0)  # No Null detection
        self.set_ordering_flag(s)

        MPI.COMM_WORLD.Barrier()
        if not gui.restore_fac:
            if not self.silent:
                dprint1("job1", debug.format_memory_usage())
            s.set_job(1)
            s.run()
            info1 = s.get_info(1)

            if info1 < 0:
                assert False, "MUMPS call (job1) failed. Check error log"
            if info1 > 0:
                dprint1("MUMPS return warning", info1)
            if not self.silent:
                dprint1("job2", debug.format_memory_usage())
            s.set_icntl(13, 0)
            s.set_icntl(5, 0)

            s.set_job(2)
            s.run()
            info1 = s.get_info(1)
            if info1 < 0:
                assert False, "MUMPS call (job2) failed. Check error log"
            if info1 > 0:
                dprint1("MUMPS return warning", info1)
            if gui.write_fac:
                if not self.silent:
                    dprint1("job7 (save)")
                prefix, path = self.split_dir_prefix(gui.factor_path)
                s.set_saveparam(prefix, path)
                s.set_oocparam("ooc_", path)

                # wait here to make sure path is created.
                MPI.COMM_WORLD.Barrier()

                s.set_job(7)
                s.run()
                info1 = s.get_info(1)
                if info1 != 0:
                    assert False, "MUMPS call (job7) failed. Check error log"

        else:
            if not self.silent:
                dprint1("job8 (restore)")
            pathes = gui.factor_path.split(',')
            prefix, path = self.split_dir_prefix(pathes[ifactor].strip())

            s.set_saveparam(prefix, path)
            s.set_oocparam("ooc_", path)

            MPI.COMM_WORLD.Barrier()  # wait here to make sure path is created.
            # s.set_job(-1)
            # s.run()
            s.set_job(8)
            s.run()
            info1 = s.get_info(1)
            if info1 != 0:
                assert False, "MUMPS call (job8) failed. Check error log"

    def write_inverse(self, s, b):
        if use_parallel:
            from mpi4py import MPI
        else:
            from petram.helper.dummy_mpi import MPI

        myid = MPI.COMM_WORLD.rank
        smyid = '{:0>6d}'.format(myid)
        mpi_size = MPI.COMM_WORLD.size

        datatype = self._data_type()

        nrhs = b.shape[0] if myid == 0 else 0
        if mpi_size > 1:
            nrhs = MPI.COMM_WORLD.bcast(nrhs)

        if myid == 0:
            dprint1("nrhs", nrhs)
            s.set_lrhs_nrhs(b.shape[0], nrhs)
            bb = np.zeros((b.shape[0], nrhs), dtype=datatype)
            for i in range(nrhs):
                bb[i, i] = 1.0
            bstack = np.hstack(np.transpose(bb))
            bstack = self.make_vector_entries(bstack)
            s.set_rhs(self.data_array(bstack))

        lsol_loc, isol_loc, sol_loc = self.set_distributed_sol(s, nrhs)
        self.set_error_analysis(s)

        # set to distributed sol mode
        use_distributed_save = True

        s.set_job(3)
        s.run()
        info1 = s.get_info(1)
        if info1 != 0:
            assert False, "MUMPS call (job3) failed. Check error log"

        #sol = s.get_sol_loc()
        # if use_distributed_save:
        sol = np.transpose(sol_loc.reshape(-1, lsol_loc))
        np.savez("matrix_inv_idx." + smyid, A_inv_idx=isol_loc)
        np.savez("matrix_inv." + smyid, A_inv=sol)

        s.set_icntl(21, 0)

    def Mult(self, b, x=None, case_base=0):
        if use_parallel:
            from mpi4py import MPI
        else:
            from petram.helper.dummy_mpi import MPI

        from petram.ext.mumps.mumps_solve import i_array

        myid = MPI.COMM_WORLD.rank
        nproc = MPI.COMM_WORLD.size

        gui = self.gui
        s = self.s
        if gui.write_mat:
            if myid == 0:
                for ib in range(b.shape[1]):
                    write_vector('rhs_' + str(ib + case_base), b[:, ib])
                case_base = case_base + b.shape[1]
            else:
                case_base = None
            case_base = MPI.COMM_WORLD.bcast(case_base, root=0)

        if self.gui.write_inv:
            self.write_inverse(s, b)

        nrhs = b.shape[1] if b is not None else 0
        if nproc > 1:
            nrhs = MPI.COMM_WORLD.bcast(nrhs)

        distributed_rhs = (self.gui.use_dist_rhs and
                           use_parallel and
                           nproc > 1)

        if self.skip_solve:
            assert not distributed_rhs, "distributed RHS is not supported for skip solve"
            print("skip solve is on, returning zero")
            return b*0.

        # this one keep it from crashing if a user choose
        # distributed RHS with std_solver.
        # For now, we don't do this but add a note in GUI panel.
        # distributed_rhs = (self.gui.use_dist_rhs and
        #                   use_parallel and
        #                   nproc > 1 and
        #                   self.keep_sol_distributed )
        if distributed_rhs:
            s.set_icntl(20, 10)

            rhs_loc = b.flatten()
            rhs_loc = rhs_loc.astype(self._data_type(), copy=False)
            irhs_loc_read = self.irhs_loc.astype(dtype=self._int_type())

            '''
            if myid == 0:
                irhs_loc_read = irhs_loc_read.astype(dtype=self._int_type())
            else:
                irhs_loc_read = irhs_loc_read.astype(dtype=self._int_type())
            '''
            #print("irhc_loc from mfem", irhs_loc)
            rhs_locc = self.data_array(rhs_loc)
            irhs_locc = i_array(irhs_loc_read)

            lrhs_loc = len(rhs_loc)
            nloc_rhs = lrhs_loc * nrhs

            s.set_nrhs_lrhs_irhs_rhs_loc(
                nloc_rhs, lrhs_loc, irhs_locc, rhs_locc)

        else:
            s.set_icntl(20, 0)
            if myid == 0:
                s.set_lrhs_nrhs(b.shape[0], b.shape[1])
                bstack = np.hstack(np.transpose(b))
                bstack = self.make_vector_entries(bstack)
                s.set_rhs(self.data_array(bstack))

        distributed_sol = (use_parallel and
                           nproc > 1 and
                           (self.keep_sol_distributed or self.gui.use_dist_sol))

        if distributed_sol:
            lsol_loc, isol_loc, sol_loc = self.set_distributed_sol(s, nrhs)

        else:
            if distributed_rhs and myid == 0:
                rhs = np.empty(self.N_global, dtype=self._data_type())
                s.set_rhs(self.data_array(rhs))

        if self.gui.icntl10.lower() != 'default':       # iterative refinement
            s.set_icntl(10, convert2int(self.gui.icntl10))

        if self.gui.cntl2.lower() != 'default':
            # stopping criterion for iterative refinement
            s.set_cntl(2, convert2float(self.gui.cntl2))

        if self.silent or self.gui.log_level >= 10:
            s.set_icntl(1, -1)
            s.set_icntl(2, -1)
            s.set_icntl(3, -1)
            s.set_icntl(4, 0)
        else:
            dprint1("job3", debug.format_memory_usage())

        if not distributed_rhs and not distributed_sol:
            self.set_error_analysis(s)

        s.set_job(3)
        s.run()

        info1 = s.get_info(1)
        info2 = s.get_info(2)
        if info1 < 0:
            assert False, "MUMPS call (job3) failed. Check error log (info1/info2)" + str(
                info1) + " " + str(info2)
        if info1 > 0:
            dprint1("MUMPS call (job3) return warning Check error log (info1/info2)" + str(
                    info1) + " " + str(info2))

        rsol = None
        isol = None
        sol_extra = None

        if distributed_sol:
            sol = self.redistributed_array(
                sol_loc.flatten(), isol_loc, self.irhs_loc, nrhs)

            if self.gui.use_dist_sol:
                sol = np.transpose(sol.reshape(nrhs, -1))
            else:
                # preconditioner mode. keep sol as it is
                pass
        else:
            if myid == 0:
                if self.is_complex:
                    sol = s.get_real_rhs() + 1j * s.get_imag_rhs()
                else:
                    sol = s.get_real_rhs()

                sol = np.transpose(sol.reshape(-1, len(b)))
            else:
                sol = None

        #nicePrint('doen with dist_RHS, dist_sol', distributed_rhs, distributed_sol)

        return sol


if use_parallel:
    import mfem.par as mfem
    from mfem.common.mpi_debug import nicePrint

else:
    import mfem.ser as mfem
    nicePrint = dprint1


class MUMPSPreconditioner(mfem.PyOperator):
    def __init__(self, A0, gui=None, engine=None, silent=True, **kwargs):
        mfem.PyOperator.__init__(self, A0.Height())
        self.gui = gui
        self.engine = engine
        self.silent = silent
        self.irhs_loc = None
        self.is_complex_operator = False

        if 'single' in kwargs and not 'double' in kwargs:
            self.single = kwargs.pop('single')
            self.gui.use_single_precision = self.single
        elif not 'single' in kwargs and 'double' in kwargs:
            self.single = not kwargs.pop('double')
            self.gui.use_single_precision = self.single
        elif 'single' in kwargs and 'double' in kwargs:
            assert False, "singel and double can not be uset together"
        else:
            pass

        self.SetOperator(A0)

    def SetOperator(self, opr):
        def isSparseMatrix(opr):
            return isinstance(opr, mfem.SparseMatrix)

        check = opr._real_operator if isinstance(
            opr, mfem.ComplexOperator) else opr

        if isSparseMatrix(check):
            from mfem.common.sparse_utils import sparsemat_to_scipycsr

            if isinstance(opr, mfem.ComplexOperator):
                mat = (sparsemat_to_scipycsr(opr._real_operator, float) +
                       sparsemat_to_scipycsr(opr._imag_operator, float) * 1j)
                coo_opr = mat.tocoo()
                self.is_complex_operator = True
            else:
                coo_opr = sparsemat_to_scipycsr(opr, float).tocoo()

            self.solver = MUMPSSolver(self.gui, self.engine)
            self.solver.AllocSolver(self.is_complex_operator,
                                    self.gui.use_single_precision)
            self.solver.SetOperator(coo_opr, False)
            self.solver.keep_sol_distributed = True
            self.row_part = [-1, -1]

        else:
            from mfem.common.parcsr_extra import ToScipyCoo
            from scipy.sparse import coo_matrix

            if isinstance(opr, mfem.ComplexOperator):
                lcsr = (ToScipyCoo(opr._real_operator).tocsr() +
                        ToScipyCoo(opr._imag_operator).tocsr() * 1j)
                lcoo = lcsr.tocoo()
                shape = (opr._real_operator.GetGlobalNumRows(),
                         opr._real_operator.GetGlobalNumCols())

                rpart = opr._real_operator.GetRowPartArray()
                self.is_complex_operator = True
                self.row_part = rpart
            else:
                lcoo = ToScipyCoo(opr)
                shape = (opr.GetGlobalNumRows(), opr.GetGlobalNumCols())
                rpart = opr.GetRowPartArray()
                self.row_part = rpart

            gcoo = coo_matrix(shape)
            gcoo.data = lcoo.data
            gcoo.row = lcoo.row + rpart[0]
            gcoo.col = lcoo.col
            self.solver = MUMPSSolver(self.gui, self.engine)
            self.solver.AllocSolver(self.is_complex_operator,
                                    self.gui.use_single_precision)
            self.solver.SetOperator(gcoo, True)
            self.solver.keep_sol_distributed = True
            self.is_parallel = True

        self.solver.set_silent(self.silent)

    def Mult(self, x, y):
        # in the parallel enviroment, we need to collect x and
        # redistribute y
        # we keep RowPart array from opr since here y is
        # vector not ParVector even in the parallel env.
        if use_parallel:
            from mpi4py import MPI
        else:
            from petram.helper.dummy_mpi import MPI

        myid = MPI.COMM_WORLD.rank
        nproc = MPI.COMM_WORLD.size

        if self.is_complex_operator:
            vec = x.GetDataArray()
            ll = vec.size
            vec = vec[:ll // 2] + 1j * vec[ll // 2:]
        else:
            vec = x.GetDataArray()

        if self.row_part[0] == -1:
            xx = np.atleast_2d(vec).transpose()
        else:
            from mpi4py import MPI
            comm = MPI.COMM_WORLD
            from petram.helper.mpi_recipes import gather_vector
            if self.gui.use_dist_rhs:
                xx = np.atleast_2d(vec).transpose()
            else:
                xx = gather_vector(vec)
                if myid == 0:
                    xx = np.atleast_2d(xx).transpose()

        # if myid == 0:
        #    print("xx shape (at node-0)", xx.shape)

        s = self.solver.Mult(xx)

        if self.row_part[0] != -1:
            distributed_sol = use_parallel and nproc > 1
            if not distributed_sol:
                comm = MPI.COMM_WORLD
                s = comm.bcast(s)
                s = s[self.row_part[0]:self.row_part[1]]

        if self.is_complex_operator:
            s = np.hstack((s.real.flatten(), s.imag.flatten()))

        #nicePrint(s.shape, y.Size())
        y.Assign(s.flatten().astype(float, copy=False))


def check_block_operator(A):
    from petram.solver.strumpack_model import get_block

    rows = A.NumRowBlocks()
    cols = A.NumColBlocks()

    is_complex = False
    is_parallel = False
    for i in range(rows):
        for j in range(cols):
            m = get_block(A, i, j)
            if m is None:
                continue
            if isinstance(m, mfem.ComplexOperator):
                is_complex = True
                check = m._real_operator
            else:
                check = m
            if not isinstance(check, mfem.SparseMatrix):
                is_parallel = True

    return is_complex, is_parallel


class MUMPSBlockPreconditioner(mfem.Solver):
    def __init__(self, opr, gui=None, engine=None, silent=False, **kwargs):
        self.gui = gui
        self.engine = engine
        self.silent = silent
        self.is_complex_operator = False
        if opr is not None:
            self._opr = weakref.ref(opr)
        else:
            self._opr = None
        self.irhs_loc = None
        super(MUMPSBlockPreconditioner, self).__init__()

    def real_to_complex(self, x):
        '''
        if use_parallel:
           from mpi4py import MPI
           myid     = MPI.COMM_WORLD.rank


           offset = M.RowOffsets().ToList()
           of = [np.sum(MPI.COMM_WORLD.allgather(np.int32(o))) for o in offset]
           if myid != 0: return

        else:
            pass
        '''

        rows = len(self.block_size)
        of = self.block_offset
        pt = 0
        x2 = np.zeros(of[-1] // 2, dtype=self.dtype)

        for i in range(rows):
            l = of[i + 1] - of[i]
            w = int(l // 2)
            x2[pt:pt + w] = x[of[i]:of[i] + w] + 1j * x[(of[i] + w):of[i + 1]]
            pt = pt + w

        return x2

    def complex_to_real(self, y):
        '''
        if use_parallel:
           from mpi4py import MPI
           myid     = MPI.COMM_WORLD.rank


           offset = M.RowOffsets().ToList()
           of = [np.sum(MPI.COMM_WORLD.allgather(np.int32(o))) for o in offset]
           if myid != 0: return

        else:
            pass
        '''
        rows = len(self.block_size)
        of = self.block_offset
        pt = 0
        y2 = np.zeros(of[-1], dtype=np.float)
        for i in range(rows):
            l = of[i + 1] - of[i]
            w = int(l // 2)
            y2[of[i]:(of[i] + w)] = y[pt:pt + w].real
            y2[(of[i] + w):(of[i] + w + w)] = y[pt:pt + w].imag
            pt = pt + w

        #assert False, "test"
        return y2

    def Mult(self, x, y):
        if use_parallel:
            from petram.helper.mpi_recipes import (gather_vector,
                                                   scatter_vector,
                                                   allgather_vector)
            from mpi4py import MPI
        else:
            from petram.helper.dummy_mpi import MPI
        myid = MPI.COMM_WORLD.rank
        nproc = MPI.COMM_WORLD.size

        if self.is_complex_operator:
            vec = x.GetDataArray()
            vec = self.real_to_complex(vec)
        else:
            vec = x.GetDataArray()

        if self.row_offset == -1:
            xx = np.atleast_2d(vec).transpose()
        else:
            if self.gui.use_dist_rhs:
                xx = np.atleast_2d(vec).transpose()
            else:
                xx = gather_vector(vec)
                if myid == 0:
                    xx = np.atleast_2d(xx).transpose()

        # if myid == 0:
        #    print("xx shape (at node-0)", xx.shape)

        s = [solver.Mult(xx) for solver in self.solver]

        if self.row_offset != -1:
            # if myid == 0:
            #w = [0.8*np.exp(1j*77/180*np.pi), np.exp(-1j*60/180*np.pi)*1.2]
            w = [1]*len(s)
            s = [xx.flatten()*w[i] for i, xx in enumerate(s)]
            s = np.mean(s, 0)
            # else:
            #    s = None
            distributed_sol = use_parallel and nproc > 1
            if not distributed_sol:
                size = np.sum(self.all_block_size, 0)[myid]
                s = scatter_vector(s, rcounts=size)

            if self.is_complex_operator:
                s = self.complex_to_real(s)

        else:
            s = [xx.flatten() for xx in s]
            s = np.mean(s, 0)
            if self.is_complex_operator:
                s = self.complex_to_real(s)

        y.Assign(s.flatten().astype(float, copy=False))

    def SetOperator(self, opr):
        if use_parallel:
            from mpi4py import MPI
            from petram.helper.mpi_recipes import (gather_vector,
                                                   allgather_vector)
            nprc = MPI.COMM_WORLD.size
            myid = MPI.COMM_WORLD.rank
            smyid = '{:0>6d}'.format(myid)

        #opr = mfem.Opr2BlockOpr(opr)
        #self._opr = weakref.ref(opr)

        opr = self._opr()
        from petram.solver.strumpack_model import build_csr_local

        is_complex, is_parallel = check_block_operator(opr)

        if is_complex:
            if self.gui.use_single_precision:
                dtype = np.complex64
            else:
                dtype = np.complex128
        else:
            if self.gui.use_single_precision:
                dtype = np.float32
            else:
                dtype = np.float64
        self.dtype = dtype
        self.is_complex_operator = is_complex

        lcsr = build_csr_local(opr, dtype, is_complex)

        if use_parallel:
            from scipy.sparse import coo_matrix

            lcoo = lcsr.tocoo()
            rows = MPI.COMM_WORLD.allgather(lcoo.shape[0])
            global_offsets = np.hstack((0, np.cumsum(rows)))
            self.row_offset = global_offsets[myid]

            shape = (np.sum(rows), lcoo.shape[1])
            gcoo = coo_matrix(shape)
            gcoo.data = lcoo.data
            gcoo.row = lcoo.row + self.row_offset
            gcoo.col = lcoo.col
        else:
            gcoo = lcsr.tocoo()
            self.row_offset = -1

        self.block_offset = np.array(opr.RowOffsets().ToList())
        self.block_size = self.block_offset[1:] - self.block_offset[:-1]

        if use_parallel:
            # create this to distibute solution vector
            self.all_block_size = allgather_vector(self.block_size)
            self.all_block_size = self.all_block_size.reshape(
                nprc, -1).transpose()
            if is_complex:
                self.all_block_size //= 2
            self.mpi_block_size = np.sum(
                self.all_block_size, 0)   # size for each mpi

        else:
            self.irhs_loc = None

        self.solver = []

        if not self.gui.restore_fac:
            solver = MUMPSSolver(self.gui, self.engine)
            solver.AllocSolver(self.is_complex_operator,
                               self.gui.use_single_precision)
            solver.SetOperator(gcoo, is_parallel)
            solver.keep_sol_distributed = True
            self.solver.append(solver)
        else:
            pathes = self.gui.factor_path.split(',')
            for i in range(len(pathes)):
                solver = MUMPSSolver(self.gui, self.engine)
                solver.AllocSolver(self.is_complex_operator,
                                   self.gui.use_single_precision)
                solver.SetOperator(gcoo, is_parallel, ifactor=i)
                solver.keep_sol_distributed = True
                self.solver.append(solver)

        self.is_parallel = is_parallel

        if is_parallel:
            pass
        else:
            self.row_part = [-1, -1]

        for x in self.solver:
            x.set_silent(self.silent)


class MUMPSBlockSolver(LinearSolver):
    '''
    MUMPS LinearSolverInstance which internally uses BlockSolver

    This is used from DerivedValue solver in TimeDependent solver.

    Not recommended to use in future, since it collect solution in
    root node.
    '''
    is_iterative = True

    def __init__(self, *args, **kwargs):
        super(MUMPSBlockSolver, self).__init__(*args, **kwargs)
        self.silent = False
        self.keep_sol_distributed = False

    def SetOperator(self, A, dist, name=None, ifactor=0):
        solver = MUMPSBlockPreconditioner(A,
                                          gui=self.gui,
                                          engine=self.engine,)
        solver.SetOperator(A)
        self._solver = solver

    def Mult(self, b, x=None, case_base=0):
        self._solver.Mult(b[0], x)

        if use_parallel:
            from mpi4py import MPI
            from petram.helper.mpi_recipes import (gather_vector,
                                                   allgather_vector)
            myid = MPI.COMM_WORLD.rank

            xx = gather_vector(x.GetDataArray())
            if myid == 0:
                xx = np.atleast_2d(xx).transpose()
        else:
            xx = x.GetDataArray().copy().reshape(-1, 1)

        return xx



from __future__ import print_function
'''
    partial_mesh


    generate a new mesh which contains only spedified domain/boundaries.

    we want to assign boundary attribute number consistently.

    MFEM currently does not have a mechanism to assign numbers
    to ndim-2 and below elements.
    We take this informatnion from extended_connectivity data
    gathered when loading mesh. This way, edge(vertex) numbers
    in 3D (2D) mesh is properied carrid over to surface. 
'''
import os
import numpy as np

from petram.mfem_config import use_parallel
import petram.debug
dprint1, dprint2, dprint3 = petram.debug.init_dprints('partial_mesh')

partition_method = "default"   # default or 0 (0 handles disconnected mesh)

if use_parallel:
    import mfem.par as mfem

    from mpi4py import MPI
    myid = MPI.COMM_WORLD.rank
    nprc = MPI.COMM_WORLD.size
    comm = MPI.COMM_WORLD

    smyid = '{:0>6d}'.format(myid)
    from mfem.common.mpi_debug import nicePrint, niceCall
    from petram.helper.mpi_recipes import allgather, allgather_vector, gather_vector
    from petram.mesh.mesh_utils import distribute_shared_entity
else:
    import mfem.ser as mfem
    myid = 0

    def nicePrint(x):
        print(x)


def isParMesh(mesh):
    return hasattr(mesh, 'GetNGroups')


def _collect_data(index, mesh, mode, skip_vtx=False):
    '''
    collect  index : attribute

    return  idx, attrs, ivert, nverts, base
      idx, attrs: element index, attr number for elemnt
      ivert : flattened vertex
      nverts : num of vertices for each element
      base : element geometry base
    '''
    LEN = len
    if mode == 'bdr':
        GetXElementVertices = mesh.GetBdrElementVertices
        GetXBaseGeometry = mesh.GetBdrElementBaseGeometry
        attrs = mesh.GetBdrAttributeArray()
        idx = np.arange(len(attrs))[np.in1d(attrs, index)]
        attrs = attrs[idx]

    elif mode == 'dom':
        GetXElementVertices = mesh.GetElementVertices
        GetXBaseGeometry = mesh.GetElementBaseGeometry
        attrs = mesh.GetAttributeArray()

        idx = np.arange(len(attrs))[np.in1d(attrs, index)]
        attrs = attrs[idx]

    elif mode == 'edge':
        GetXElementVertices = mesh.GetEdgeVertices
        def GetXBaseGeometry(x): return 1

        s2l = mesh.extended_connectivity['surf2line']
        l2e = mesh.extended_connectivity['line2edge']
        idx = np.array(sum([l2e[ea] for ea in index], []), dtype=int)
        if len(idx) == 0:
            attrs = np.atleast_1d([]).astype(int)
        else:
            attrs = np.hstack([[ea]*len(l2e[ea]) for ea in index]).astype(int)

    elif mode == 'vertex':
        v2v = mesh.extended_connectivity['vert2vert']
        def GetXElementVertices(x): return v2v[x] if x in v2v else []
        def GetXBaseGeometry(x): return 0
        def LEN(x): return 1
        idx = np.array([x for x in index if x in v2v], dtype=int)
        if len(idx) == 0:
            attrs = np.atleast_1d([]).astype(int)
        else:
            attrs = np.hstack([va for va in index if va in v2v]).astype(int)

    else:
        assert False, "Unknown mode (_collect_data) "+mode

    if len(idx) > 0:
        ivert = [GetXElementVertices(i) for i in idx]
        nverts = np.array([LEN(x) for x in ivert], dtype=int)
        ivert = np.hstack(ivert).astype(int, copy=False)
        base = np.hstack([GetXBaseGeometry(i)
                          for i in idx]).astype(int, copy=False)
    else:
        ivert = np.array([], dtype=int)
        nverts = np.array([], dtype=int)
        base = np.array([], dtype=int)

    return idx, attrs, ivert, nverts, base


def _add_face_data(m, idx, nverts, base):
    '''
    3D mesh (face), 2D mesh (edge)
    '''
    new_v = [m.GetFaceVertices(i) for i in idx]
    #ivert  = np.hstack((ivert,  np.hstack(new_v)))
    nverts = np.hstack((nverts,  [len(kk) for kk in new_v]))
    base = np.hstack((base,  [m.GetFaceBaseGeometry(i) for i in idx]))
    # return ivert, nverts, base
    return nverts, base


def _gather_shared_vertex(mesh, u, shared_info,  *iverts):
    # u_own, iv1, iv2... = gather_shared_vertex(mesh, u, ld, md, iv1, iv2...)

    # u_own  : unique vertex id ownd by a process
    # shared_info : shared data infomation
    # iv1, iv2, ...: array of vertex is after overwriting shadow vertex
    #                to a real one, which is owned by other process

    # process shared vertex
    #    1) a vertex in sub-volume may be shadow
    #    2) the real one may not a part of sub-volume on the master node
    #    3) we always use the real vertex
    #        1) shadow vertex index is over-written to a real vertex
    #        2) make sure that the real one is added to sub-volume mesh obj.

    offset = np.hstack([0, np.cumsum(allgather(mesh.GetNV()))])
    iverts = [iv + offset[myid] for iv in iverts]
    u = u + offset[myid]  # -> global numbering

    ld, md = shared_info
    mv_list = [[] for i in range(nprc)]
    for key in ld.keys():
        mid, g_in_master = key
        if mid != myid:
            for lv, mv in zip(ld[key][0], md[key][0]):
                ic = 0
                for iv in iverts:
                    iii = np.where(iv == lv)[0]
                    ic = ic + len(iii)
                    if len(iii) > 0:
                        iv[iii] = mv
                if ic > 0:
                    mv_list[mid].append(mv)
            u = u[np.in1d(u, ld[key][0], invert=True)]
    for i in range(nprc):
        mvv = gather_vector(np.atleast_1d(mv_list[i]).astype(int), root=i)
        if i == myid:
            missing = np.unique(mvv[np.in1d(mvv, u, invert=True)])
            if len(missing) != 0:
                dprint1("adding (vertex)", missing)
                u = np.hstack((u, missing))

    u_own = np.sort(u - offset[myid])
    return [u_own]+list(iverts)  # u_own, iv1, iv2 =


def _gather_shared_element(mesh, mode, shared_info, ielem, kelem, attrs,
                           nverts, base, ivert, skip_adding=False):

    ld, md = shared_info

    if mode == 'face':
        imode = 2
    elif mode == 'edge':
        imode = 1
    else:
        imode = 0
    #
    me_list = [[] for i in range(nprc)]
    mea_list = [[] for i in range(nprc)]
    for key in ld.keys():
        mid, g_in_master = key
        if mid != myid:
            for le, me in zip(ld[key][imode], md[key][imode]):
                iii = np.where(ielem == le)[0]
                if len(iii) != 0:
                    if not skip_adding:
                        kelem[iii] = False
                    me_list[mid].append(me)
                    mea_list[mid].extend(list(attrs[iii]))
                assert len(
                    iii) < 2, "same iface (pls report this error to developer) ???"

    new_ivert = np.array([], dtype=ivert.dtype)
    for i in range(nprc):
        mev = gather_vector(np.atleast_1d(me_list[i]).astype(int), root=i)
        mea = gather_vector(np.atleast_1d(mea_list[i]).astype(int), root=i)
        if i == myid:
            check = np.in1d(mev, ielem, invert=True)
            missing, mii = np.unique(mev[check], return_index=True)
            missinga = mea[check][mii]
            if len(missing) != 0:
                dprint1("adding (face)", myid, missing)
                dprint1(len(missing), len(missinga), missinga)
                if not skip_adding:
                    nverts, base = _add_face_data(mesh, missing, nverts, base)
                    attrs = np.hstack((attrs, missinga)).astype(attrs.dtype)
                    kelem = np.hstack((kelem, [True]*len(missing)))

    if not skip_adding:
        new_ivert = new_ivert.astype(ivert.dtype)
        new_ivert = allgather_vector(new_ivert)
        #ivert  = np.hstack((ivert,  new_ivert))
    attrs = allgather_vector(attrs)
    base = allgather_vector(base)
    nverts = allgather_vector(nverts)
    kelem = allgather_vector(kelem)
    return kelem, attrs, nverts, base, ivert


def _fill_mesh_elements(omesh, vtx, indices, nverts,  attrs, base):

    cnverts = np.hstack([0, np.cumsum(nverts)])

    for i, a in enumerate(attrs):
        iv = indices[cnverts[i]:cnverts[i+1]]
        if base[i] == 1:  # segment
            el = mfem.Segment()
            el.SetAttribute(a)
            el.SetVertices(list(iv))
            el.thisown = False
            omesh.AddElement(el)
        elif base[i] == 2:  # triangle
            omesh.AddTri(list(iv), a)
        elif base[i] == 3:  # quad
            omesh.AddQuad(list(iv), a)
        elif base[i] == 4:  # tet
            omesh.AddTet(list(iv), a)
        elif base[i] == 5:  # hex
            omesh.AddHex(list(iv), a)
        else:
            assert False, "unsupported base geometry: " + str(base[i])


def _fill_mesh_bdr_elements(omesh, vtx, bindices, nbverts,
                            battrs, bbase, kbelem):

    cnbverts = np.hstack([0, np.cumsum(nbverts)])
    for i, ba in enumerate(battrs):
        if not kbelem[i]:
            dprint1("skipping")
            continue
        iv = bindices[cnbverts[i]:cnbverts[i+1]]
        if bbase[i] == 0:
            el = mfem.Point(iv[0])
            el.SetAttribute(ba)
            el.thisown = False
            omesh.AddBdrElement(el)
        elif bbase[i] == 1:
            omesh.AddBdrSegment(list(iv), ba)
        elif bbase[i] == 2:
            #if myid == 1: print(list(iv), ba)
            omesh.AddBdrTriangle(list(iv), ba)
        elif bbase[i] == 3:
            omesh.AddBdrQuad(list(iv), ba)
        else:
            assert False, "unsupported base geometry: " + str(bbase[i])

    for v in vtx:
        omesh.AddVertex(list(v))


def edge(mesh, in_attr, filename='', precision=8):
    '''
    make a new mesh which contains only spedified edges.

    in_attr : eihter
    filename : an option to save the file 
    return new surface mesh
    '''
    sdim = mesh.SpaceDimension()
    dim = mesh.Dimension()
    Nodal = mesh.GetNodalFESpace()
    hasNodal = (Nodal is not None)

    if sdim == 3 and dim == 3:
        mode = 'edge', 'vertex'
    elif sdim == 3 and dim == 2:
        mode = 'bdr', 'vertex'
    elif sdim == 2 and dim == 2:
        mode = 'bdr', 'vertex'
    elif sdim == 2 and dim == 1:
        mode = 'dom', 'vertex'
    else:
        assert False, "unsupported mdoe"

    idx, attrs, ivert, nverts, base = _collect_data(in_attr, mesh, mode[0])

    l2v = mesh.extended_connectivity['line2vert']
    in_eattr = np.unique(np.hstack([l2v[k]
                         for k in in_attr])).astype(int, copy=False)
    if isParMesh(mesh):
        in_eattr = np.unique(allgather_vector(in_eattr))
    eidx, eattrs, eivert, neverts, ebase = _collect_data(in_eattr, mesh,
                                                         mode[1])

    u, indices = np.unique(np.hstack((ivert, eivert)),
                           return_inverse=True)
    keelem = np.array([True]*len(eidx), dtype=bool)
    u_own = u

    if isParMesh(mesh):
        shared_info = distribute_shared_entity(mesh)
        u_own, ivert, eivert = _gather_shared_vertex(mesh, u, shared_info,
                                                     ivert, eivert)
    Nvert = len(u)
    if len(u_own) > 0:
        vtx = np.vstack([mesh.GetVertexArray(i) for i in u_own])
    else:
        vtx = np.array([]).reshape((-1, sdim))

    if isParMesh(mesh):
        #
        # distribute vertex/element data
        #
        base = allgather_vector(base)
        nverts = allgather_vector(nverts)
        attrs = allgather_vector(attrs)

        ivert = allgather_vector(ivert)
        eivert = allgather_vector(eivert)

        vtx = allgather_vector(vtx.flatten()).reshape(-1, sdim)

        u, indices = np.unique(np.hstack([ivert, eivert]),
                               return_inverse=True)

        #
        # take care of shared boundary (edge)
        #
        keelem, eattrs, neverts, ebase, eivert = (
            _gather_shared_element(mesh, 'vertex', shared_info, eidx,
                                   keelem, eattrs,
                                   neverts, ebase, eivert))

    indices = np.array([np.where(u == biv)[0][0] for biv in ivert])
    eindices = np.array([np.where(u == biv)[0][0] for biv in eivert])

    Nvert = len(vtx)
    Nelem = len(attrs)
    Nbelem = len(eattrs)

    dprint1("NV, NBE, NE: " +
            ",".join([str(x) for x in (Nvert, Nbelem, Nelem)]))

    omesh = mfem.Mesh(1, Nvert, Nelem, Nbelem, sdim)

    _fill_mesh_elements(omesh, vtx, indices, nverts, attrs, base)
    _fill_mesh_bdr_elements(omesh, vtx, eindices, neverts, eattrs,
                            ebase, keelem)

    omesh.FinalizeTopology()

    if hasNodal:
        odim = omesh.Dimension()

        dprint1("odim, dim, sdim", odim, " ", dim, " ", sdim)
        fec = Nodal.FEColl()
        dNodal = mfem.FiniteElementSpace(omesh, fec, sdim)
        omesh.SetNodalFESpace(dNodal)
        omesh._nodal = dNodal

        GetXDofs = Nodal.GetElementDofs
        if dim == 3:
            GetXDofs = Nodal.GetEdgeDofs
        elif dim == 2:
            GetXDofs = Nodal.GetBdrElementDofs
        elif dim == 1:
            GetXDofs = Nodal.GetElementDofs

        dGetXDofs = dNodal.GetElementDofs

        DofToVDof = Nodal.DofToVDof
        dDofToVDof = dNodal.DofToVDof

        #nicePrint(dGetNX(),',', GetNX())
        nodes = mesh.GetNodes()
        node_ptx1 = nodes.GetDataArray()

        onodes = omesh.GetNodes()
        node_ptx2 = onodes.GetDataArray()
        #nicePrint(len(idx), idx)

        if len(idx) > 0:
            dof1_idx = np.hstack([[DofToVDof(i, d) for d in range(sdim)]
                                  for j in idx
                                  for i in GetXDofs(j)])
            data = node_ptx1[dof1_idx]
        else:
            dof1_idx = np.array([])
            data = np.array([])
        if isParMesh(mesh):
            data = allgather_vector(data)
        if isParMesh(mesh):
            idx = allgather_vector(idx)
        #nicePrint(len(data), ',', len(idx))

        dof2_idx = np.hstack([[dDofToVDof(i, d) for d in range(sdim)]
                              for j in range(len(idx))
                              for i in dGetXDofs(j)])
        node_ptx2[dof2_idx] = data
        # nicePrint(len(dof2_idx))

    # this should be after setting HO nodals...
    omesh.Finalize(refine=True, fix_orientation=True)

    if isParMesh(mesh):
        if omesh.GetNE() < nprc*3:
            parts = omesh.GeneratePartitioning(1, 1)
        elif partition_method != 'default':
            nparts = min([nprc, 8])
            parts = omesh.GeneratePartitioning(nparts, 0)
        else:
            parts = None
        omesh = mfem.ParMesh(comm, omesh, parts)

    if filename != '':
        if isParMesh(mesh):
            smyid = '{:0>6d}'.format(myid)
            filename = filename + '.'+smyid
        omesh.PrintToFile(filename, precision)

    return omesh


def surface(mesh, in_attr, filename='', precision=8):
    '''
    mesh must be 
    if sdim == 3:
       a domain of   2D mesh
       a boundary of 3D mesh
    if sdim == 2:
       a domain  in 2D mesh

    in_attr : eihter
    filename : an option to save the file 
    return new surface mesh

    '''
    sdim = mesh.SpaceDimension()
    dim = mesh.Dimension()
    Nodal = mesh.GetNodalFESpace()
    hasNodal = (Nodal is not None)

    if sdim == 3 and dim == 3:
        mode = 'bdr', 'edge'
    elif sdim == 3 and dim == 2:
        mode = 'dom', 'bdr'
    elif sdim == 2 and dim == 2:
        mode = 'dom', 'bdr'
    else:
        assert False, "unsupported mdoe"

    idx, attrs, ivert, nverts, base = _collect_data(in_attr, mesh, mode[0])

    s2l = mesh.extended_connectivity['surf2line']
    in_eattr = np.unique(np.hstack([s2l[k]
                         for k in in_attr])).astype(int, copy=False)
    if isParMesh(mesh):
        in_eattr = np.unique(allgather_vector(in_eattr))

    eidx, eattrs, eivert, neverts, ebase = _collect_data(in_eattr, mesh,
                                                         mode[1])

    u, indices = np.unique(np.hstack((ivert, eivert)),
                           return_inverse=True)
    keelem = np.array([True]*len(eidx), dtype=bool)
    u_own = u

    if isParMesh(mesh):
        shared_info = distribute_shared_entity(mesh)
        u_own, ivert, eivert = _gather_shared_vertex(mesh, u, shared_info,
                                                     ivert, eivert)
    Nvert = len(u)
    if len(u_own) > 0:
        vtx = np.vstack([mesh.GetVertexArray(i) for i in u_own])
    else:
        vtx = np.array([]).reshape((-1, sdim))

    if isParMesh(mesh):
        #
        # distribute vertex/element data
        #
        base = allgather_vector(base)
        nverts = allgather_vector(nverts)
        attrs = allgather_vector(attrs)

        ivert = allgather_vector(ivert)
        eivert = allgather_vector(eivert)

        vtx = allgather_vector(vtx.flatten()).reshape(-1, sdim)

        u, indices = np.unique(np.hstack([ivert, eivert]),
                               return_inverse=True)

        #
        # take care of shared boundary (edge)
        #
        keelem, eattrs, neverts, ebase, eivert = (
            _gather_shared_element(mesh, 'edge', shared_info, eidx,
                                   keelem, eattrs,
                                   neverts, ebase, eivert,
                                   skip_adding=True))

    #indices  = np.array([np.where(u == biv)[0][0] for biv in ivert])
    #eindices = np.array([np.where(u == biv)[0][0] for biv in eivert])

    iv, ivi = np.unique(ivert, return_inverse=True)
    tmp = np.where(np.in1d(u,  ivert,  assume_unique=True))[0]
    indices = tmp[ivi]
    iv, ivi = np.unique(eivert, return_inverse=True)
    tmp = np.where(np.in1d(u,  eivert, assume_unique=True))[0]
    eindices = tmp[ivi]

    Nvert = len(vtx)
    Nelem = len(attrs)
    Nbelem = len(eattrs)

    dprint1("NV, NBE, NE: " +
            ",".join([str(x) for x in (Nvert, Nbelem, Nelem)]))

    omesh = mfem.Mesh(2, Nvert, Nelem, Nbelem, sdim)

    _fill_mesh_elements(omesh, vtx, indices, nverts, attrs, base)
    _fill_mesh_bdr_elements(omesh, vtx, eindices, neverts, eattrs,
                            ebase, keelem)

    omesh.FinalizeTopology()

    if hasNodal:
        odim = omesh.Dimension()
        dprint1("odim, dim, sdim", odim, " ", dim, " ", sdim)
        fec = Nodal.FEColl()
        dNodal = mfem.FiniteElementSpace(omesh, fec, sdim)
        omesh.SetNodalFESpace(dNodal)
        omesh._nodal = dNodal

        if sdim == 3:
            if dim == 3:
                GetXDofs = Nodal.GetBdrElementDofs
                GetNX = Nodal.GetNBE
            elif dim == 2:
                GetXDofs = Nodal.GetElementDofs
                GetNX = Nodal.GetNE
            else:
                assert False, "not supported ndim 1"
        elif sdim == 2:
            GetNX = Nodal.GetNE
            GetXDofs = Nodal.GetElementDofs

        dGetNX = dNodal.GetNE
        dGetXDofs = dNodal.GetElementDofs

        DofToVDof = Nodal.DofToVDof
        dDofToVDof = dNodal.DofToVDof

        #nicePrint(dGetNX(),',', GetNX())
        nodes = mesh.GetNodes()
        node_ptx1 = nodes.GetDataArray()

        onodes = omesh.GetNodes()
        node_ptx2 = onodes.GetDataArray()
        #nicePrint(len(idx), idx)

        if len(idx) > 0:
            dof1_idx = np.hstack([[DofToVDof(i, d) for d in range(sdim)]
                                  for j in idx
                                  for i in GetXDofs(j)])
            data = node_ptx1[dof1_idx]
        else:
            dof1_idx = np.array([])
            data = np.array([])
        if isParMesh(mesh):
            data = allgather_vector(data)
        if isParMesh(mesh):
            idx = allgather_vector(idx)
        #nicePrint(len(data), ',', len(idx))

        dof2_idx = np.hstack([[dDofToVDof(i, d) for d in range(sdim)]
                              for j in range(len(idx))
                              for i in dGetXDofs(j)])
        node_ptx2[dof2_idx] = data

    # this should be after setting HO nodals...
    omesh.Finalize(refine=True, fix_orientation=True)

    if isParMesh(mesh):
        if omesh.GetNE() < nprc*3:
            parts = omesh.GeneratePartitioning(1, 1)
        elif partition_method != 'default':
            nparts = min([nprc, 8])
            parts = omesh.GeneratePartitioning(nparts, 0)
        else:
            parts = None
        omesh = mfem.ParMesh(comm, omesh, parts)
        #omesh = mfem.ParMesh(comm, omesh)

    if filename != '':
        if isParMesh(mesh):
            smyid = '{:0>6d}'.format(myid)
            filename = filename + '.'+smyid
        omesh.PrintToFile(filename, precision)

    return omesh


def volume(mesh, in_attr, filename='', precision=8):
    '''
    make a new mesh which contains only spedified attributes.

    note: 
       1) boundary elements are also copied and bdr_attributes
          are maintained
       2) in parallel, new mesh must be geometrically continuous.
          this routine does not check it

    mesh must have sdim == 3:
    in_attr : domain attribute
    filename : an option to save the file 

    return new volume mesh
    '''
    in_attr = np.atleast_1d(in_attr)
    sdim = mesh.SpaceDimension()
    dim = mesh.Dimension()
    Nodal = mesh.GetNodalFESpace()
    hasNodal = (Nodal is not None)

    if sdim != 3:
        assert False, "sdim must be three for volume mesh"
    if dim != 3:
        assert False, "sdim must be three for volume mesh"

    idx, attrs, ivert, nverts, base = _collect_data(in_attr, mesh, 'dom')

    v2s = mesh.extended_connectivity['vol2surf']
    in_battr = np.unique(np.hstack([v2s[k]
                         for k in in_attr])).astype(int, copy=False)
    if isParMesh(mesh):
        in_battr = np.unique(allgather_vector(in_battr))

    bidx, battrs, bivert, nbverts, bbase = _collect_data(in_battr, mesh, 'bdr')
    iface = np.array([mesh.GetBdrElementEdgeIndex(i) for i in bidx],
                     dtype=int)

    # note u is sorted unique
    u, indices = np.unique(np.hstack((ivert, bivert)),
                           return_inverse=True)

    kbelem = np.array([True]*len(bidx), dtype=bool)
    u_own = u

    if isParMesh(mesh):
        shared_info = distribute_shared_entity(mesh)
        u_own, ivert, bivert = _gather_shared_vertex(mesh, u, shared_info,
                                                     ivert, bivert)

    if len(u_own) > 0:
        vtx = np.vstack([mesh.GetVertexArray(i) for i in u_own])
    else:
        vtx = np.array([]).reshape((-1, sdim))

    if isParMesh(mesh):
        #
        # distribute vertex/element data
        #
        base = allgather_vector(base)
        nverts = allgather_vector(nverts)
        attrs = allgather_vector(attrs)

        ivert = allgather_vector(ivert)
        bivert = allgather_vector(bivert)

        vtx = allgather_vector(vtx.flatten()).reshape(-1, sdim)

        u, indices = np.unique(np.hstack([ivert, bivert]), return_inverse=True)

        #
        # take care of shared boundary (face)
        #
        #  2018.11.28
        #   skip_adding is on. This basically skip shared_element
        #   processing. Check em3d_TEwg7 if you need to remov this.
        #
        kbelem, battrs, nbverts, bbase, bivert = (
            _gather_shared_element(mesh, 'face', shared_info, iface,
                                   kbelem, battrs,
                                   nbverts, bbase, bivert,
                                   skip_adding=True))

    #indices0  = np.array([np.where(u == biv)[0][0] for biv in ivert])
    #bindices0 = np.array([np.where(u == biv)[0][0] for biv in bivert])

    iv, ivi = np.unique(ivert, return_inverse=True)
    tmp = np.where(np.in1d(u,  ivert,  assume_unique=True))[0]
    indices = tmp[ivi]
    iv, ivi = np.unique(bivert, return_inverse=True)
    tmp = np.where(np.in1d(u,  bivert, assume_unique=True))[0]
    bindices = tmp[ivi]

    #print('check', np.sum(np.abs(indices - indices0)))

    Nvert = len(vtx)
    Nelem = len(attrs)
    Nbelem = np.sum(kbelem)  # len(battrs)

    dprint1("NV, NBE, NE: " +
            ",".join([str(x) for x in (Nvert, Nbelem, Nelem)]))

    omesh = mfem.Mesh(3, Nvert, Nelem, Nbelem, sdim)
    #omesh = mfem.Mesh(3, Nvert, Nelem, 0, sdim)

    _fill_mesh_elements(omesh, vtx, indices, nverts, attrs, base)
    _fill_mesh_bdr_elements(omesh, vtx, bindices, nbverts, battrs,
                            bbase, kbelem)

    omesh.FinalizeTopology()

    if hasNodal:
        odim = omesh.Dimension()
        dprint1("odim, dim, sdim", odim, " ", dim, " ", sdim)

        fec = Nodal.FEColl()
        dNodal = mfem.FiniteElementSpace(omesh, fec, sdim)
        omesh.SetNodalFESpace(dNodal)
        omesh._nodal = dNodal

        GetXDofs = Nodal.GetElementDofs
        GetNX = Nodal.GetNE
        dGetXDofs = dNodal.GetElementDofs
        dGetNX = dNodal.GetNE

        DofToVDof = Nodal.DofToVDof
        dDofToVDof = dNodal.DofToVDof

        #nicePrint(dGetNX(),',', GetNX())
        nodes = mesh.GetNodes()
        node_ptx1 = nodes.GetDataArray()

        onodes = omesh.GetNodes()
        node_ptx2 = onodes.GetDataArray()
        #nicePrint(len(idx), idx)

        if len(idx) > 0:
            dof1_idx = np.hstack([[DofToVDof(i, d) for d in range(sdim)]
                                  for j in idx
                                  for i in GetXDofs(j)])
            data = node_ptx1[dof1_idx]
        else:
            dof1_idx = np.array([])
            data = np.array([])
        if isParMesh(mesh):
            data = allgather_vector(data)
        if isParMesh(mesh):
            idx = allgather_vector(idx)
        #nicePrint(len(data), ',', len(idx))

        dof2_idx = np.hstack([[dDofToVDof(i, d) for d in range(sdim)]
                              for j in range(len(idx))
                              for i in dGetXDofs(j)])
        node_ptx2[dof2_idx] = data

    # this should be after setting HO nodals...
    omesh.Finalize(refine=True, fix_orientation=True)

    if isParMesh(mesh):
        if omesh.GetNE() < nprc*3:
            parts = omesh.GeneratePartitioning(1, 1)
        elif partition_method != 'default':
            nparts = min([nprc, 8])
            parts = omesh.GeneratePartitioning(nparts, 0)
        else:
            parts = None
        omesh = mfem.ParMesh(comm, omesh, parts)
        #omesh = mfem.ParMesh(comm, omesh)

    if filename != '':
        if isParMesh(mesh):
            smyid = '{:0>6d}'.format(myid)
            filename = filename + '.'+smyid
        omesh.PrintToFile(filename, precision)

    return omesh



import numpy as np


def expand_vertex_data(X, vertex_idx, element_id):
    '''
    expand index data using element_id, so that

       surface edges will have duplicate vertex,
       this way,,
         1) the normal vector on the edge become discutinous
         2) the elment index on the surface becomes constant
    '''
    k = 0
    verts = []
    iele = []
    iarr = []

    nel = vertex_idx.shape[-1]

    def iter_unique_idx(x):
        '''
        return unique element and all indices for each eleement

        note: this is to avoid using where in the loop like this
           # for kk in np.unique(element_id):
           #    idx = np.where(element_id == kk)[0]
           # (using where inside the loop makes it very slow as
           # number of unique elements increases
        '''
        uv, uc = np.unique(x, return_counts=True)
        idx = np.argsort(x)
        uc = np.hstack((0, np.cumsum(uc)))
        for i, v in enumerate(uv):
            yield (v, idx[uc[i]:uc[i + 1]])

    for kk, idx in iter_unique_idx(element_id):
        iverts = vertex_idx[idx].flatten()
        iv, idx = np.unique(iverts, return_inverse=True)

        verts.append(X[iv])
        iele.append(idx.reshape(-1, nel) + k)
        k = k + len(iv)
        iarr.append(np.zeros(len(iv)) + kk)

    array_idx = np.hstack(iarr).astype(int)
    elem_idx = np.vstack(iele).astype(int)
    verts = np.vstack(verts)

    return verts, elem_idx, array_idx


# we visit nodes in counter clock wise.
gmsh_ho_el_idx = {'triangle6': ([0, 3, 5], [3, 1, 4], [5, 3, 4], [5, 4, 2]),
                  'triangle10': ([0, 3, 8], [8, 3, 9], [3, 4, 9], [9, 4, 5], [1, 5, 4],
                                 [7, 8, 9], [7, 9, 6], [6, 9, 5], [2, 7, 6]),
                  'triangle15': ([0, 3, 11], [11, 3, 12], [12, 3, 4], [4, 13, 12], [13, 4, 5],
                                 [13, 5, 6], [6, 5, 1], [
                                     10, 11, 12], [10, 12, 14],
                                 [14, 12, 13], [14, 13, 7], [
                                     7, 13, 6], [9, 10, 14],
                                 [9, 14, 8], [8, 14, 7], [2, 9, 8]),
                  'triangle21': ([0, 3, 14], [3, 15, 14], [3, 4, 15], [4, 18, 15],
                                 [4, 5, 18], [5, 16, 18], [
                                     5, 6, 16], [6, 7, 16],
                                 [6, 1, 7], [13, 14, 15], [13, 15, 20],
                                 [15, 18, 20], [18, 19, 20], [
                                     18, 16, 19], [16, 8, 19],
                                 [16, 7, 8], [12, 13, 20], [
                                     20, 17, 12], [20, 19, 17],
                                 [19, 9, 17], [19, 8, 9], [
                                     11, 12, 17], [11, 17, 10],
                                 [10, 17, 9], [2, 11, 10]),

                  'quad9': ([0, 4, 7], [7, 4, 8], [8, 4, 1], [8, 1, 5], [3, 7, 8],
                            [3, 8, 6], [6, 8, 5], [6, 5, 2],),
                  'quad16': ([0, 4, 11], [11, 4, 12], [4, 5, 12], [12, 5, 13], [5, 1, 13],
                             [13, 1, 6], [11, 12, 10], [10, 12, 15], [
                                 12, 13, 15], [15, 13, 14],
                             [13, 6, 14], [14, 6, 7], [10, 15, 3], [
                                 3, 15, 9], [15, 14, 9],
                             [9, 14, 8], [14, 7, 8], [8, 7, 2]),
                  'quad25': ([0, 4, 15], [15, 4, 16], [4, 5, 16], [16, 5, 20], [5, 6, 20],
                             [20, 6, 17], [6, 1, 17], [17, 1, 7], [
                                 15, 16, 14], [14, 16, 23],
                             [16, 20, 23], [23, 20, 24], [
                                 20, 17, 24], [24, 17, 21], [17, 7, 21],
                             [21, 7, 8], [14, 23, 13], [13, 23, 19], [
                                 23, 24, 19], [19, 24, 22],
                             [24, 21, 22], [22, 21, 18], [
                                 21, 8, 18], [18, 8, 9], [13, 19, 3],
                             [3, 19, 12], [19, 22, 12], [
                                 12, 22, 11], [22, 18, 11],
                             [11, 18, 10], [18, 9, 10], [10, 9, 2]),
                  'quad36': ([0, 4, 19], [19, 4, 20], [4, 5, 20], [20, 5, 24], [5, 6, 24],
                             [24, 6, 25], [6, 7, 25], [
                                 25, 7, 21], [7, 1, 21], [21, 1, 8],
                             [19, 20, 18], [18, 20, 31], [20, 24, 31], [
                                 31, 24, 32], [24, 25, 32],
                             [32, 25, 33], [25, 21, 33], [
                                 33, 21, 26], [21, 8, 26], [26, 8, 9],
                             [18, 31, 17], [17, 31, 30], [31, 32, 30], [
                                 30, 32, 35], [32, 33, 35],
                             [35, 33, 34], [33, 26, 34], [
                                 34, 26, 27], [26, 9, 27], [27, 9, 10],
                             [17, 30, 16], [16, 30, 23], [30, 35, 23], [
                                 23, 35, 29], [35, 34, 29],
                             [29, 34, 28], [34, 27, 28], [28, 27, 22], [
                                 27, 10, 22], [22, 10, 11],
                             [16, 23, 3], [3, 23, 15], [23, 29, 15], [
                                 15, 29, 14], [29, 28, 14],
                             [14, 28, 13], [28, 22, 13], [13, 22, 12], [22, 11, 12], [12, 11, 2]),
                  }

gmsh_ho_bel_idx = {'triangle6': ([0, 3], [3, 1], [1, 4], [4, 2], [2, 5], [5, 0]),
                   'triangle10': ([0, 3], [3, 4], [4, 1],
                                  [1, 5], [5, 6], [6, 2],
                                  [2, 7], [7, 8], [8, 0]),
                   'triangle15': ([0, 3], [3, 4], [4, 5], [5, 1],
                                  [1, 6], [6, 7], [7, 8], [8, 2],
                                  [2, 9], [9, 10], [10, 11], [11, 0]),
                   'triangle21': ([0, 3], [3, 4], [4, 5], [5, 6], [6, 1],
                                  [1, 7], [7, 8], [8, 9], [9, 10], [10, 2],
                                  [2, 11], [11, 12], [12, 13], [13, 14], [14, 0]),
                   'quad9': ([0, 4], [4, 1], [1, 5], [5, 2], [2, 6], [6, 3],
                             [3, 7], [7, 0]),
                   'quad16': ([0, 4], [4, 5], [5, 1], [1, 6], [6, 7], [7, 2],
                              [2, 8], [8, 9], [9, 3], [3, 10], [10, 11], [11, 0]),
                   'quad25': ([0, 4], [4, 5], [5, 6], [6, 1],
                              [1, 7], [7, 8], [8, 9], [9, 2],
                              [2, 10], [10, 11], [11, 12], [12, 3],
                              [3, 13], [13, 14], [14, 15], [15, 0]),
                   'quad36': ([0, 4], [4, 5], [5, 6], [6, 7], [7, 1],
                              [1, 8], [8, 9], [8, 10], [10, 11], [11, 2],
                              [2, 12], [12, 13], [13, 14], [14, 15], [15, 3],
                              [3, 16], [16, 17], [17, 18], [18, 19], [19, 0],)}


def expand_vertex_data_gmsh_ho(X, vertex_idx, element_id, el_type):
    '''
    expand index data using element_id, so that

       surface edges will have duplicate vertex,
       this way,,
         1) the normal vector on the edge become discutinous
         2) the elment index on the surface becomes constant
    '''
    k = 0
    verts = []
    iele = []
    iarr = []
    iedge = []

    nel = vertex_idx.shape[-1]

    def iter_unique_idx(x):
        '''
        return unique element and all indices for each eleement

        note: this is to avoid using where in the loop like this
           # for kk in np.unique(element_id):
           #    idx = np.where(element_id == kk)[0]
           # (using where inside the loop makes it very slow as
           # number of unique elements increases
        '''
        uv, uc = np.unique(x, return_counts=True)
        idx = np.argsort(x)
        uc = np.hstack((0, np.cumsum(uc)))
        for i, v in enumerate(uv):
            yield (v, idx[uc[i]:uc[i + 1]])

    for kk, idx in iter_unique_idx(element_id):
        iverts = vertex_idx[idx].flatten()
        iv, idx = np.unique(iverts, return_inverse=True)

        verts.append(X[iv])
        tmp = idx.reshape(-1, nel) + k

        el = [tmp[:, x] for x in gmsh_ho_el_idx[el_type]]
        bel = [tmp[:, x] for x in gmsh_ho_bel_idx[el_type]]
        elem_idx = np.vstack(el)
        ed_idx = np.vstack(bel)

        iele.append(elem_idx)
        iedge.append(ed_idx)
        k = k + len(iv)
        iarr.append(np.zeros(len(iv)) + kk)

    array_idx = np.hstack(iarr).astype(int)
    elem_idx = np.vstack(iele).astype(int)
    edge_idx = np.vstack(iedge).astype(int)
    verts = np.vstack(verts)

    return verts, elem_idx, edge_idx, array_idx


def call_solid1(viewer, name, verts, elem_idx, array_idx, lw,
                edge_idx=None):
    # template for faces
    obj = viewer.solid(verts, elem_idx,
                       array_idx=array_idx,
                       facecolor=(0.7, 0.7, 0.7, 1.0),
                       linewidth=lw,
                       edge_idx=edge_idx)

    obj.rename(name)
    obj.set_gl_hl_use_array_idx(True)
    return obj


def call_solid2(viewer, name, verts, elem_idx, array_idx=None, lw=1.5):
    # template for lines
    obj = viewer.solid(verts, elem_idx,
                       array_idx=array_idx,
                       linewidth=lw,
                       facecolor=(0, 0, 0, 1.0),
                       edgecolor=(0, 0, 0, 1.0),
                       #                           view_offset = (0, 0, -0.001, 0),
                       draw_last=True)
    obj.rename(name)
    obj.set_gl_hl_use_array_idx(True)
    return obj


def plot_geometry(viewer, ret, geo_phys='geometrical', lw=0):
    '''
    plot_geometry
    '''
    viewer.cls()
    viewer.set_hl_color((1, 0, 0))

    X, cells, pt_data, cell_data, field_data = ret

    if 'triangle_x' in cells or 'quad_x' in cells:
        # merge two cases as triangle data
        verts = cell_data['X_refined_face']

        elem_idx0 = []
        eelem_idx0 = []
        array_idx = np.zeros(len(verts), dtype=int)

        for n in ['triangle_x', 'quad_x']:
            if not n in cells:
                continue
            xxx = cells[n]
            if n == 'quad_x':
                elem_idx0.append(xxx[:, :3])
                elem_idx0.append(xxx[:, [0, 2, 3]])
            else:
                elem_idx0.append(xxx)
            eelem_idx0.append(cells[n + 'e'])
            idx = cell_data[n][geo_phys]
            array_idx[np.unique(cells[n])] = idx

        elem_idx = np.vstack(elem_idx0)
        eelem_idx = np.vstack(eelem_idx0)

        call_solid1(viewer, 'face_t', verts, elem_idx, array_idx, lw,
                    edge_idx=eelem_idx)

    elif 'triangle' in cells or 'quad' in cells:
        verts0 = []
        elem_idx0 = []
        eelem_idx0 = []
        array_idx0 = []

        l_verts = 0
        for n in ['triangle', 'quad']:
            if not n in cells:
                continue
            verts, elem_idx, array_idx = expand_vertex_data(X, cells[n],
                                                            cell_data[n][geo_phys])
            verts0.append(verts)
            xxx = elem_idx if l_verts == 0 else elem_idx + l_verts

            if n == 'quad':
                elem_idx0.append(xxx[:, :3])
                elem_idx0.append(xxx[:, [0, 2, 3]])
                eelem_idx0.extend([xxx[:, :2],
                                   xxx[:, 1:3],
                                   xxx[:, -2:],
                                   xxx[:, [0, -1]]])
            else:
                elem_idx0.append(xxx)
                eelem_idx0.extend([xxx[:, :2],
                                   xxx[:, -2:],
                                   xxx[:, [0, 2]]])
            l_verts = l_verts + len(verts)
            array_idx0.append(array_idx)

        verts = np.vstack(verts0)
        elem_idx = np.vstack(elem_idx0)
        array_idx = np.hstack(array_idx0)
        eelem_idx = np.vstack(eelem_idx0)

        call_solid1(viewer, 'face_t', verts, elem_idx, array_idx, lw,
                    edge_idx=eelem_idx)

    '''
    if 'triangle_x' in cells:
        elem_idx = cells['triangle_x']
        array_idx = cell_data['triangle_x'][geo_phys]
        eelem_idx = cells['triangle_xe']

        array_idx = np.zeros(len(verts))


        array_idx[np.unique(elem_idx.flatten()] =

        ### to handle mixed mesh, where verts contains vertices for tri and quad
        ar, idx = np.unique(np.hstack([elem_idx.flatten(),
                                       eelem_idx.flatten()]), return_inverse=True)
        elem_idx = idx[:np.prod(elem_idx.shape)].reshape(elem_idx.shape)
        eelem_idx = idx[np.prod(elem_idx.shape):].reshape(eelem_idx.shape)
        verts = verts[ar]


        call_solid1(viewer, 'face_t', verts, elem_idx, array_idx, lw,
                    edge_idx=eelem_idx)
    if 'quad_x' in cells:
        verts = cell_data['X_refined_face']
        elem_idx = cells['quad_x']
        array_idx = cell_data['quad_x'][geo_phys]
        eelem_idx = cells['quad_xe']

        ### to handle mixed mesh, where verts contains vertices for tri and quad
        ar, idx = np.unique(np.hstack([elem_idx.flatten(),
                                       eelem_idx.flatten()]), return_inverse=True)
        elem_idx = idx[:np.prod(elem_idx.shape)].reshape(elem_idx.shape)
        eelem_idx = idx[np.prod(elem_idx.shape):].reshape(eelem_idx.shape)
        verts = verts[ar]

        call_solid1(viewer, 'face_r', verts, elem_idx, array_idx, lw,
                    edge_idx=eelem_idx)
    '''

    if 'line_x' in cells:
        verts = cell_data['X_refined_edge']
        elem_idx = cells['line_x']
        array_idx = cell_data['line_x'][geo_phys]
        call_solid2(viewer, 'edge', verts, elem_idx, array_idx)

    elif 'line' in cells and len(cells['line']) > 0:
        verts, elem_idx, array_idx = expand_vertex_data(X, cells['line'],
                                                        cell_data['line'][geo_phys])
        call_solid2(viewer, 'edge', verts, elem_idx, array_idx)

    if 'vertex' in cells:
        if 'vertex_mask' in cells:
            vidx = cells['vertex'][cells['vertex_mask']]
            aidx = cell_data['vertex'][geo_phys][cells['vertex_mask']]
        else:
            vidx = cells['vertex']
            aidx = cell_data['vertex'][geo_phys]
        if len(vidx) > 0:
            vert = np.atleast_2d(np.squeeze(X[vidx]))
        #if len(vert) > 0:
            x = vert[:, 0]
            if vert.shape[-1] > 1:
                y = vert[:, 1]
            else:
                y = x*0
            if vert.shape[-1] > 2:
                z = vert[:, 2]
            else:
                z = y*0
            obj = viewer.plot(x, y, z, 'ok',
                              array_idx=aidx,
                              linewidth=0)
            obj.rename('point')
            obj.set_gl_hl_use_array_idx(True)
    viewer.set_sel_mode(viewer.get_sel_mode())


def oplot_meshed(viewer, ret):
    print("oplot_meshed")
    ax = viewer.get_axes()
    for name, obj in ax.get_children():
        if name.endswith('_meshed'):
            viewer.cls(obj=obj)

    try:
        X, cells, pt_data, cell_data, field_data = ret
    except ValueError:
        return

    meshed_face = []
    if 'triangle' in cells:
        verts, elem_idx, array_idx = expand_vertex_data(X, cells['triangle'],
                                                        cell_data['triangle']['geometrical'])

        obj = viewer.solid(verts, elem_idx,
                           array_idx=array_idx,
                           facecolor=(0.7, 0.7, 0.7, 1.0),
                           edgecolor=(0, 0, 0, 1),
                           linewidth=1,
                           view_offset=(0, 0, -0.0005, 0))

        obj.rename('face_t_meshed')
        obj.set_gl_hl_use_array_idx(True)

        meshed_face.extend(
            list(
                np.unique(
                    cell_data['triangle']['geometrical'])))

    def handle_gmsh_ho_elements(el):
        elem_idx = cells[el]
        array_idx = cell_data[el]['geometrical']
        verts, elem_idx, edge_idx, array_idx = expand_vertex_data_gmsh_ho(
            X, elem_idx, array_idx, el)

        obj = viewer.solid(verts, elem_idx,
                           array_idx=array_idx,
                           facecolor=(0.7, 0.7, 0.7, 1.0),
                           edgecolor=(0, 0, 0, 1),
                           linewidth=1,
                           view_offset=(0, 0, -0.0005, 0),
                           edge_idx=edge_idx)
        obj.rename('face_t_meshed')
        obj.set_gl_hl_use_array_idx(True)

        meshed_face.extend(list(array_idx))
    if 'triangle6' in cells:
        handle_gmsh_ho_elements('triangle6')
    if 'triangle10' in cells:
        handle_gmsh_ho_elements('triangle10')
    if 'triangle15' in cells:
        handle_gmsh_ho_elements('triangle15')
    if 'triangle21' in cells:
        handle_gmsh_ho_elements('triangle21')

    if 'quad' in cells:
        verts, elem_idx, array_idx = expand_vertex_data(X, cells['quad'],
                                                        cell_data['quad']['geometrical'])

        # print verts.shape, elem_idx.shape, array_idx.shape
        obj = viewer.solid(verts, elem_idx,
                           array_idx=array_idx,
                           facecolor=(0.7, 0.7, 0.7, 1.0),
                           edgecolor=(0, 0, 0, 1),
                           #                           view_offset = (0, 0, -0.0005, 0),
                           linewidth=1,)

        obj.rename('face_r_meshed')
        obj.set_gl_hl_use_array_idx(True)
        meshed_face.extend(list(np.unique(cell_data['quad']['geometrical'])))

    if 'quad9' in cells:
        handle_gmsh_ho_elements('quad9')
    if 'quad16' in cells:
        handle_gmsh_ho_elements('quad16')
    if 'quad25' in cells:
        handle_gmsh_ho_elements('quad25')
    if 'quad36' in cells:
        handle_gmsh_ho_elements('quad36')

    hide_face_meshmode(viewer, meshed_face)
    '''
    s, v = viewer._s_v_loop['mesh']
    facesa = []
    if len(v)>0:  # in 3D starts with faces from shown volumes
        all_surfaces = np.array(list(s), dtype=int)
        for key in v:
            if not key in viewer._mhidden_volume:
                facesa.extend(v[key])
        facesa = np.unique(facesa)
        mask  = np.logical_not(np.in1d(all_surfaces, facesa))
        facesa = list(all_surfaces[mask])

    facesa.extend(viewer._mhidden_face)

    for name, obj in ax.get_children():
        if name.startswith('face') and not name.endswith('meshed'):
            h = list(np.unique(facesa + meshed_face))
            obj.hide_component(h)
        if name.startswith('face') and  name.endswith('meshed'):
            h = list(np.unique(facesa))
            obj.hide_component(h)
    '''
    if 'line' in cells:
        vert = np.squeeze(X[cells['line']][:, 0, :])
        if vert.size > 3:
            obj = viewer.plot(vert[:, 0],
                              vert[:, 1],
                              vert[:, 2], 'ob',
                              array_idx=cell_data['line']['geometrical'],
                              linewidth=0)
#                    view_offset = (0, 0, -0.005, 0))

            verts, elem_idx, array_idx = expand_vertex_data(X, cells['line'],
                                                            cell_data['line']['geometrical'])

            obj.rename('edge_meshed')
            meshed_edge = list(np.unique(cell_data['line']['geometrical']))
    else:
        meshed_edge = []

    hide_edge_meshmode(viewer, meshed_edge)
    '''
    s, v = viewer._s_v_loop['mesh']
    edgesa = viewer._mhidden_edge

    for name, obj in ax.get_children():
        if name.startswith('edge') and not name.endswith('meshed'):
            h = list(np.unique(edgesa + meshed_edge))
            if obj.hasvar('idxset'): obj.hide_component(h)
    '''
    viewer.set_sel_mode(viewer.get_sel_mode())


def hide_face_meshmode(viewer, meshed_face):
    ax = viewer.get_axes()
    s, v = viewer._s_v_loop['mesh']
    facesa = []
    if v is not None and len(
            v) > 0:  # in 3D starts with faces from shown volumes
        all_surfaces = np.array(list(s), dtype=int)
        for key in v:
            if not key in viewer._mhidden_volume:
                facesa.extend(v[key])
        facesa = np.unique(facesa)
        mask = np.logical_not(np.in1d(all_surfaces, facesa))
        facesa = list(all_surfaces[mask])

    facesa.extend(viewer._mhidden_face)

    for name, obj in ax.get_children():
        if name.startswith('face') and not name.endswith('meshed'):
            h = list(np.unique(facesa + meshed_face))
            obj.hide_component(h)
        if name.startswith('face') and name.endswith('meshed'):
            h = list(np.unique(facesa))
            obj.hide_component(h)


def hide_edge_meshmode(viewer, meshed_edge):
    ax = viewer.get_axes()
    s, v = viewer._s_v_loop['mesh']
    edgesa = viewer._mhidden_edge

    for name, obj in ax.get_children():
        if name.startswith('edge') and not name.endswith('meshed'):
            h = list(np.unique(edgesa + meshed_edge))
            if obj.hasvar('idxset'):
                obj.hide_component(h)



from petram.phys.weakform import get_integrators
import numpy as np

import petram
import petram.phys
from os import listdir
from os.path import dirname, basename, isfile, join
import glob
import warnings
import glob
import sys

# collect all physics module

_modulenames = []
for p in petram.phys.__path__:
    mm = glob.glob(join(p, "*", "__init__.py"))
    for item in [basename(dirname(m)) for m in mm]:
        if item == 'coeff2d':  # we don't use this anymore
            continue
        if not item in _modulenames:
            _modulenames.append(item)


def all_phys_models():
    modulenames = []
    for m in _modulenames:
        if m == "common":  # this is the name for common physics subroutines
            continue
        try:
            mname = 'petram.phys.'+m + '.'+m+'_model'
            __import__(mname, locals(), globals())

            mm = getattr(petram.phys, m)
            mmm = getattr(mm, m+'_model')
            invalid = hasattr(mmm, "dependency_invalid")

            if invalid:
                import petram.mfem_model as mm
                if mm.has_addon_access == "any" or mm.has_addon_access == m:
                    modulenames.append(m)
            else:
                modulenames.append(m)
        except ImportError:
            warnings.warn('Failed to import physcis module :' + mname)

    models = []
    classes = []
    for m in modulenames:
        mm = getattr(petram.phys, m)
        models.append(getattr(mm, m+'_model'))

        chk = [x.isdigit() for x in m]
        if True in chk:
            chk = chk.index(True)
            if hasattr(models[-1], 'model_basename'):
                bs = getattr(models[-1], 'model_basename')
            else:
                bs = m[:chk+2].upper()
            classname = bs + m[chk+2:].lower()
            classes.append(getattr(models[-1], classname))
        else:
            bs = getattr(models[-1], 'model_basename')
            classes.append(getattr(models[-1], bs))
    return models, classes


def get_phys_constraints(module):
    mname = 'petram.phys.'+module
    if mname not in sys.modules:
        try:
            __import__(mname, locals(), globals())
        except ImportError:
            warnings.warn('Failed to import physcis module :' + mname)
            raise
    try:

        paths = getattr(sys.modules[mname], '__path__')
    except BaseException:
        warnings.warn('Can not find path for :' + mname)
        raise

    constraints = {'domain': [], 'bdry': [],
                   'edge': [], 'pointt': [], 'pair': []}

    for p in paths:
        for f in listdir(p):
            name = basename(f)
            if not name.endswith('py'):
                continue
            if name.startswith('_'):
                continue
            name = name[:-3]
            mname2 = mname + '.' + name

            try:
                __import__(mname2, locals(), globals())
            except BaseException:
                warnings.warn('Failed to import physcis module :' + mname2)
                raise

            for key in constraints:
                if hasattr(sys.modules[mname2], key+'_constraints'):
                    m = getattr(sys.modules[mname2], key+'_constraints')
                    for x in m():
                        if not x in constraints[key]:
                            constraints[key].append(x)

    return constraints


bilinintegs = get_integrators('BilinearOps')
linintegs = get_integrators('LinearOps')

itg_choices = bilinintegs + linintegs


def restrict_coeff(coeff, isDomain, sel_index, engine, emesh_idx,
                   vec=False, matrix=False):

    from petram.engine import mfem
    mesh = engine.emeshes[emesh_idx]

    if isDomain:
        arrsize = max(np.max(mesh.attributes.ToList()), engine.max_attr)
    else:
        arrsize = max(np.max(mesh.bdr_attributes.ToList()), engine.max_bdrattr)

    if isinstance(sel_index, str):
        sel_index = [sel_index]
    if len(sel_index) == 1 and sel_index[0] == 'all':
        return coeff

    arr = [0]*arrsize
    for k in sel_index:
        arr[k-1] = 1
    arr = mfem.intArray(arr)

    if vec:
        return mfem.VectorRestrictedCoefficient(coeff, arr)
    elif matrix:
        return mfem.MatrixRestrictedCoefficient(coeff, arr)
    else:
        return mfem.RestrictedCoefficient(coeff, arr)


def restricted_integrator(engine, integrator, sel_index,
                          c_txt, cotype,  codim,  emesh_idx,
                          isDomain, ind_vars, local_ns, global_ns, real):
    '''
    make a integrator with restricted coefficient
    (sample input)
    integrator : 'MassIntegrator'
    sel_index  : ['all'], [1,2,3]
    c_txt      : return value of self.vt_coeff.make_value_or_expression(self)[0]
    cotype     : 'S', 'V', 'M', 'D' (coefficient type)
    codim      :  3 (coefficient dim)
    ind_vars   : ('x', 'y', 'z')

    others are obvious....;D
    '''
    from petram.engine import mfem
    from petram.phys.weakform import VCoeff, SCoeff, MCoeff, DCoeff

    if cotype == 'S':
        for b in itg_choices:
            if b[0] == integrator:
                break
        if not "S*2" in b[3]:
            coeff = SCoeff(c_txt, ind_vars, local_ns, global_ns,
                           real=real)
        else:  # so far this is only for an elastic integrator
            coeff = (SCoeff(c_txt, ind_vars, local_ns, global_ns,
                            real=real, component=0),
                     SCoeff(c_txt, ind_vars, local_ns, global_ns,
                            real=real, component=1))
        coeff = restrict_coeff(coeff, isDomain, sel_index, engine, emesh_idx)
    elif cotype == 'V':
        coeff = VCoeff(codim, c_txt, ind_vars, local_ns, global_ns,
                       real=real)
        coeff = restrict_coeff(coeff, isDomain, coeff, sel_index, engine, emesh_idx,
                               vec=True)
    elif cotype == 'M':
        coeff = MCoeff(codim, c_txt, ind_vars, local_ns, global_ns,
                       real=real)
        coeff = restrict_coeff(coeff, isDomain, coeff, sel_index, engine, emesh_idx,
                               matrix=True)
    elif cotype == 'D':
        coeff = DCoeff(codim, c_txt, ind_vars, local_ns, global_ns,
                       real=real)
        coeff = restrict_coeff(coeff, isDomain, sel_index, engine, emesh_idx,
                               matrix=True)

    if coeff is None:
        return
    #if not isinstance(coeff, tuple): coeff = (coeff, )

    integrator = getattr(mfem, integrator)
    itg = integrator(coeff)
    itg._linked_coeff = coeff  # make sure that coeff is not GCed.

    return itg


def default_lf_integrator(info1, isDomain):
    if (info1['element'].startswith('ND') or
            info1['element'].startswith('RT')):
        if isDomain:
            integrator = 'VectorFEDomainLFIntegrator'
        else:
            if info1['element'].startswith('ND'):
                integrator = 'VectorFEBoundaryTangentLFIntegrator'
            else:
                integrator = 'VectorFEBoundaryFluxLFIntegrator'
    elif info1['element'].startswith('DG'):
        assert False, "auto selection is not supported for GD"
    elif info1['vdim'] > 1:
        if isDomain:
            integrator = 'VectorDomainLFIntegrator'
        else:
            integrator = 'VectorBoundaryLFIntegrator'
    elif not isDomain:
        integrator = 'BoundaryLFIntegrator'
    else:
        integrator = 'DomainLFIntegrator'
    return integrator


def default_bf_integrator(info1, info2, isDomain):
    if info1 == info2:
        if (info1['element'].startswith('ND') or
                info1['element'].startswith('RT')):
            integrator = 'VectorFEMassIntegrator'
        elif info1['element'].startswith('DG'):
            assert False, "auto selection is not supported for GD"
        elif info1['vdim'] > 1:
            integrator = 'VectorMassIntegrator'
        elif not isDomain:
            integrator = 'BoundaryMassIntegrator'
        else:
            integrator = 'MassIntegrator'
    else:
        if (((info1['element'].startswith('ND') or
              info1['element'].startswith('RT')) and
             (info2['element'].startswith('ND') or
              info2['element'].startswith('RT')))):
            integrator = 'MixedVectorMassIntegrator'
        elif info1['element'].startswith('DG'):
            assert False, "auto selection is not supported for GD"
        elif (info1['vdim'] == 1 and info1['vdim'] == info2['vdim']):
            integrator = 'MixedScalarMassIntegrator'
        else:
            assert False, "No proper integrator is found"
    return integrator



import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('MeshExtension')

from petram.mfem_config import use_parallel

if use_parallel:
   import mfem.par as mfem
   from mfem.common.mpi_debug import nicePrint
else:
   def nicePrint(x):
      print(x)

class MeshExtInfo(dict):
    

    def __init__(self, *args, **kwargs):
        cut = set(kwargs.pop('cut', []))
        sel = set(kwargs.pop('sel', []))
        dim = kwargs.pop('dim', 1)
        base = kwargs.pop('base', 0)        
        self.set_selection(sel)
        self.set_cut(cut)
        self['dim'] = dim
        self['base'] = base

        dict.__init__(self, *args, **kwargs)

    def set_dim(self, dim):
        self['dim'] = dim
        
    def set_selection(self, sel):
        self['sel']  = set([int(x) for x in sel])

    def set_cut(self, sel):
        self['cut']  = set([int(x) for x in sel])

    def __eq__(self, x):
        for key in x:
            if not key in self: return False
            if x[key] != self[key]: return False
        return True
    
class MeshExt(object):
    '''
    Mesh Manipulation 

    '''
    def __init__(self):
        from petram.model import ModelDict
        self.info = []

    def do_add_info(self, info):
        '''
        add info and return mesh index
        '''
        for k, x in enumerate(self.info):
            if x == info: return k
        self.info.append(info)
        return len(self.info)-1
    
    def add_default_info(self, j, dim, sel):
        info = MeshExtInfo(base = j, dim = dim, sel = sel)
        self.do_add_info(info)
        
    def get_info(self, idx):
        return self.info[idx]
     
    def add_info(self, info):
        '''
        add info and return mesh index
            if cut is on, make sure that there is a common no-cut mesh
        '''
        if len(info['cut']) != 0:
            info2 = MeshExtInfo(info)
            info2['cut'] = set(list())
            base = self.do_add_info(info2)
            info['base'] = base
        return self.do_add_info(info)
        
            

def generate_emesh(emeshes, info):
    '''
    apply menh manipulation defined by info to mesh

    '''
    from petram.mesh.partial_mesh import surface, volume, edge
    
    base_mesh = emeshes[info['base']]
    dprint1(info)
    if info['dim'] == 3:
        if base_mesh.Dimension() == 3:         
            alldom = set(base_mesh.extended_connectivity['vol2surf'].keys())
            if alldom == info['sel']: return base_mesh
            if len(info['sel']) == 0: return base_mesh
        dprint1("calling volume")
        m = volume(base_mesh, list(info['sel']))#, filename = 'par_part.mesh')
    elif info['dim'] == 2:
        if base_mesh.Dimension() == 2: 
            alldom = set(base_mesh.extended_connectivity['surf2line'].keys())
            if alldom == info['sel']: return base_mesh
            if len(info['sel']) == 0: return base_mesh
        dprint1("calling surface")            
        m = surface(base_mesh, list(info['sel']))#, filename = 'par_part.mesh')            
    elif info['dim'] == 1:
        if base_mesh.Dimension() == 1: 
            alldom = set(base_mesh.extended_connectivity['line2vert'].keys())
            if alldom == info['sel']: return base_mesh
            if len(info['sel']) == 0: return base_mesh
        dprint1("calling edge")                        
        m = edge(base_mesh, list(info['sel']))#, filename = 'par_part.mesh')            
    else:
        raise NotImplementedError("emesh with " +str(info["dim"]))
    return m



import os
import numpy as np

from petram.solver.iterative_model import (Iterative,
                                           IterativeSolver)
from petram.solver.solver_model import SolverInstance

from petram.model import Model
from petram.solver.solver_model import Solver, SolverInstance
from petram.solver.std_solver_model import StdSolver

from petram.mfem_config import use_parallel
if use_parallel:
    from petram.helper.mpi_recipes import *
    import mfem.par as mfem
else:
    import mfem.ser as mfem

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('MGSolver')
rprint = debug.regular_print('MGSolver')


class CoarseSolver:
    pass


class FineSolver:
    pass


class CoarseIterative(Iterative, CoarseSolver):
    @classmethod
    def fancy_menu_name(self):
        return 'Iterative'

    @classmethod
    def fancy_tree_name(self):
        return 'Iterative'

    def get_info_str(self):
        return 'Coarse'


class FineIterative(Iterative, FineSolver):
    @classmethod
    def fancy_menu_name(self):
        return 'Iterative'

    @classmethod
    def fancy_tree_name(self):
        return 'Iterative'

    def get_info_str(self):
        return 'Fine'


class Smoother(Iterative, FineSolver):
    @classmethod
    def fancy_menu_name(self):
        return 'Smoother'

    @classmethod
    def fancy_tree_name(self):
        return 'Smoother'

    def get_info_str(self):
        return 'Fine'


class MGSolver(StdSolver):
    def attribute_set(self, v):
        super(MGSolver, self).attribute_set(v)
        v["refinement_levels"] = "2"
        v["refinement_type"] = "P(order)"
        v["presmoother_count"] = "1"
        v["postsmoother_count"] = "1"
        return v

    def panel1_param(self):
        panels = super(MGSolver, self).panel1_param()
        panels.extend([["refinement type", self.refinement_type, 1,
                        {"values": ["P(order)", "H(mesh)"]}],
                       ["number of levels", "", 0, {}],
                       ["pre-smoother #count", "", 0, {}],
                       ["post-smoother #count", "", 0, {}], ])
        return panels

    def get_panel1_value(self):
        value = list(super(MGSolver, self).get_panel1_value())
        value.append(self.refinement_type)
        value.append(self.refinement_levels)
        value.append(self.presmoother_count)
        value.append(self.postsmoother_count)
        return value

    def import_panel1_value(self, v):
        super(MGSolver, self).import_panel1_value(v[:-4])
        self.refinement_type = v[-4]
        self.refinement_levels = v[-3]
        self.presmoother_count = v[-2]
        self.postsmoother_count = v[-1]

    def allocate_solver_instance(self, engine):
        if self.clear_wdir:
            engine.remove_solfiles()

        instance = StandardSolver(self, engine)
        return instance

    def get_matrix_weight(self, timestep_config):  # , timestep_weight):
        if timestep_config[0]:
            return [1, 0, 0]
        else:
            return [0, 0, 0]

    def get_num_levels(self):
        return int(self.refinement_levels)

    def set_model_level(self, klevel):
        '''
        change physcis model setting to assemble operator at
        differnet level
        '''
        return None

    def reset_model_level(self):
        '''
        revert physcis model setting to oritinal
        '''
        return None

    def get_possible_child(self):
        return FineIterative, CoarseIterative

    def get_possible_child_menu(self):
        choice = [("Coarse Lv. Solver", CoarseIterative),
                  ("!", None),
                  ("Fine Lv. Solver", FineIterative),
                  ("!", None)]
        return choice

    def create_refined_levels(self, engine, lvl):
        '''
        lvl : refined level number (1, 2, 3, ....)
              1 means "AFTER" 1 refinement
        '''
        if lvl >= int(self.refinement_levels):
            return False

        target_phys = self.get_target_phys()
        refine_type = self.refinement_type[0]  # P or H
        for phys in target_phys:
            dprint1("Adding refined level for " + phys.name())
            engine.prepare_refined_level(phys, refine_type, inc=1)

        engine.level_idx = lvl
        for phys in target_phys:
            engine.get_true_v_sizes(phys)

        return True

    @debug.use_profiler
    def run(self, engine, is_first=True, return_instance=False):
        dprint1("Entering run (is_first=", is_first, ")", self.fullpath())
        if self.clear_wdir:
            engine.remove_solfiles()

        instance = MGInstance(self, engine)
        instance.set_blk_mask()
        if return_instance:
            return instance

        instance.configure_probes(self.probe)

        if self.init_only:
            engine.sol = engine.assembled_blocks[1][0]
            instance.sol = engine.sol
        else:
            if is_first:
                instance.assemble()
                is_first = False
            instance.solve()

        instance.save_solution(ksol=0,
                               skip_mesh=False,
                               mesh_only=False,
                               save_parmesh=self.save_parmesh)
        engine.sol = instance.sol

        instance.save_probe()

        dprint1(debug.format_memory_usage())
        return is_first


class MGInstance(SolverInstance):
    def __init__(self, gui, engine):
        SolverInstance.__init__(self, gui, engine)
        self.assembled = False
        self.linearsolver = None

    @property
    def blocks(self):
        return self.engine.assembled_blocks

    def compute_A(self, M, B, X, mask_M, mask_B):
        '''
        M[0] x = B

        return A and isAnew
        '''
        return M[0], True

    def compute_rhs(self, M, B, X):
        '''
        M[0] x = B
        '''
        return B

    def do_assemble(self, inplace=True):
        engine = self.engine
        phys_target = self.get_phys()
        phys_range = self.get_phys_range()

        # use get_phys to apply essential to all phys in solvestep
        dprint1("Asembling system matrix",
                [x.name() for x in phys_target],
                [x.name() for x in phys_range])
        engine.run_verify_setting(phys_target, self.gui)
        engine.run_assemble_mat(phys_target, phys_range)
        engine.run_assemble_b(phys_target)
        engine.run_fill_X_block()

        self.engine.run_assemble_blocks(self.compute_A,
                                        self.compute_rhs,
                                        inplace=inplace)
        #A, X, RHS, Ae, B, M, names = blocks

    def assemble(self, inplace=True):
        engine = self.engine

        engine.level_idx = 0
        self.do_assemble(inplace)

        engine.level_idx = 1
        self.do_assemble(inplace)

        self.assembled = True

    def set_linearsolver_model(self):
        fine_solver = self.gui.get_active_solver(cls=FineSolver)
        if fine_solver is None:
            assert False, "Fine level solver is not chosen"
        coarse_solver = self.gui.get_active_solver(cls=CoarseSolver)
        if coarse_solver is None:
            assert False, "Coarse level solver is not chosen"

        phys_target = self.get_phys()

        self.linearsolver_models = [coarse_solver, fine_solver]
        self.phys_real = all([not p.is_complex() for p in phys_target])
        ls_type1 = coarse_solver.linear_system_type(self.gui.assemble_real,
                                                    self.phys_real)
        ls_type2 = fine_solver.linear_system_type(self.gui.assemble_real,
                                                  self.phys_real)
        if ls_type1 != ls_type2:
            assert False, "Fine/Coarse solvers must assmelbe the same linear system type"
        self.ls_type = ls_type1

    def finalize_linear_system(self, level):
        engine = self.engine

        engine.level_idx = level
        solver_model = self.linearsolver_models[level]
        # if not self.assembled:
        #    assert False, "assmeble must have been called"

        A, X, RHS, Ae, B, M, depvars = self.blocks
        mask = self.blk_mask
        engine.copy_block_mask(mask)

        depvars = [x for i, x in enumerate(depvars) if mask[0][i]]

        AA = engine.finalize_matrix(A, mask, not self.phys_real,
                                    format=self.ls_type)
        BB = engine.finalize_rhs([RHS], A, X[0], mask, not self.phys_real,
                                 format=self.ls_type)

        linearsolver = self.allocate_linearsolver(self.gui.is_complex(),
                                                  self.engine,
                                                  solver_model=solver_model)

        linearsolver.SetOperator(AA,
                                 dist=engine.is_matrix_distributed,
                                 name=depvars)

        if linearsolver.is_iterative:
            XX = engine.finalize_x(X[0], RHS, mask, not self.phys_real,
                                   format=self.ls_type)
        else:
            XX = None
        return linearsolver, BB, XX, AA

    def assemble_rhs(self):
        assert False, "assemble_rhs is not implemented"
        '''
        engine = self.engine
        phys_target = self.get_phys()
        engine.run_assemble_b(phys_target)
        B = self.engine.run_update_B_blocks()
        self.blocks[4] = B
        self.assembled = True
        '''

    def solve(self, update_operator=True):
        engine = self.engine

        ls0, _BB, _XX, _AA = self.finalize_linear_system(0)     # coarse
        ls1, BB, XX, AA = self.finalize_linear_system(1)     # fine

        class MyGMRESSolver(mfem.PyGMRESSolver):
            def __init__(self, *args, **kwags):
                mfem.PyGMRESSolver.__init__(self, *args, **kwags)

            def Mult(self, x, y):
                print('----- MyGMRESSolver.Mult')
                mfem.PyGMRESSolver.Mult(self, x, y)

            def MultTranspose(self, x, y):
                print('MyGMRESSolver.MultTranspose')
                mfem.PyGMRESSolver.Mult(self, x, y)
            # def SetOperator(self, x):
            #     mfem.GMRESSolver.SetOperator(self, x)

        solver1 = MyGMRESSolver()
        solver1.SetRelTol(ls0.gui.reltol)
        solver1.SetAbsTol(0.0)
        solver1.SetMaxIter(50)
        solver1.SetPrintLevel(ls0.gui.log_level)
        solver1.SetOperator(ls1.A)

        P_matrix = fill_prolongation_operator(engine, 0, ls1.A)
        prolongations = [P_matrix]
        #smoothers = [ls0.solver, genearate_smoother(engine, 0, ls1.A)]
        smoothers = [ls0.solver, solver1]
        operators = [ls0.A, ls1.A]

        #solall = linearsolver.Mult(BB, case_base=0)
        mg = MG(operators, smoothers, prolongations,
                presmoother_count=int(self.gui.presmoother_count),
                postsmoother_count=int(self.gui.postsmoother_count))

        # very small value
        # ls1.solver.SetPreconditioner(mg.solver)
        #solall = ls1.Mult(BB, XX)

        # transfer looks okay
        #solall0 = ls0.Mult(_BB, _XX)
        #np.save('saved_block_vector', _XX.GetDataArray())

        '''
        solall0 = ls0.Mult(_BB, _XX)
        engine.level_idx = 0
        A = engine.assembled_blocks[0]
        X = engine.assembled_blocks[1]
        solall = np.transpose(np.vstack([_XX.GetDataArray()]))        
        if not self.phys_real and self.gui.assemble_real:
            solall = self.linearsolver_models[-1].real_to_complex(solall, _AA)

        A.reformat_central_mat(solall, 0, X[0], self.blk_mask)
        self.sol = X[0]
        return True
        '''
        #

        # this seems good after removing ReorientTetMesh and
        # changed to use H refinement
        '''
        solall0 = ls0.Mult(_BB, _XX)
        P_matrix.Mult(_XX, XX)
        np.save('saved_block_vector', XX.GetDataArray())        
        engine.level_idx = 1
        A = engine.assembled_blocks[0]
        X = engine.assembled_blocks[1]
        solall = np.transpose(np.vstack([XX.GetDataArray()]))        
        if not self.phys_real and self.gui.assemble_real:
            solall = self.linearsolver_models[-1].real_to_complex(solall, AA)

        A.reformat_central_mat(solall, 0, X[0], self.blk_mask)
        self.sol = X[0]
        return True
        '''
        #print("here", type(_XX), _XX.GetDataArray().shape, XX.GetDataArray().shape)
        #solall = np.transpose(np.vstack([XX.GetDataArray()]))
        #

        # mg alone seems okay. but smoother destor
        #mg.solver.Mult(BB[0], XX)
        #solall = np.transpose(np.vstack([XX.GetDataArray()]))
        '''
        class MyPreconditioner(mfem.Solver):
            def __init__(self):
                mfem.Solver.__init__(self)

            def Mult(self, x, y):
                print('MyPreconditioner.Mult')                
                np.save('original_b', _BB[0].GetDataArray())
                np.save('original_x', x.GetDataArray())                
                P_matrix.MultTranspose(x, _BB[0])
                np.save('restricted_b', _BB[0].GetDataArray())
                ls0.Mult(_BB, _XX)
                P_matrix.Mult(_XX, y)
                #print(x, y)
                #assert False, "faile for now"

            def SetOperator(self, opr):
                pass

        class MyFGMRESSolver(mfem.FGMRESSolver):
            def Mult(self, x, y):
                print('MyFGMRESSolver.Mult')
                mfem.FGMRESSolver.Mult(self, x, y)

        prc = MyPreconditioner()
        # write solver here...
        solver = MyFGMRESSolver()
        solver.SetRelTol(1e-18)
        solver.SetAbsTol(0)
        solver.SetMaxIter(1)
        solver.SetPrintLevel(1)
        solver.SetOperator(ls1.A)
        solver.SetPreconditioner(mg.solver)
        #solver.SetPreconditioner(prc)
        np.save('initial_x', XX.GetDataArray())        
        solver.Mult(BB[0], XX)
        solall = np.transpose(np.vstack([XX.GetDataArray()]))
        '''
        #solall = ls1.Mult(BB, XX)
        '''
        solall = ls1.Mult(BB, XX)        
        mg.solver.Mult(BB[0], XX)

        '''
        np.save('saved_block_vector_input', BB[0].GetDataArray())
        if ls1.gui.maxiter > 0:
            ls1.solver.SetPreconditioner(mg.solver)
            solall = ls1.Mult(BB, XX)
        else:
            mg.solver.Mult(BB[0], XX)
            solall = np.transpose(np.vstack([XX.GetDataArray()]))

        np.save('saved_block_vector_mg', XX.GetDataArray())

        # check fine level linear system...
        #solall = ls1.Mult(BB, XX)

        '''
        solall = linearsolver.Mult(BB, x=XX, case_base=0)

        #linearsolver.SetOperator(AA, dist = engine.is_matrix_distributed)
        #solall = linearsolver.Mult(BB, case_base=0)
        '''
        if not self.phys_real and self.gui.assemble_real:
            solall = self.linearsolver_models[-1].real_to_complex(solall, AA)

        engine.level_idx = 1
        A = engine.assembled_blocks[0]
        X = engine.assembled_blocks[1]
        A.reformat_central_mat(solall, 0, X[0], self.blk_mask)

        self.sol = X[0]

        # store probe signal (use t=0.0 in std_solver)
        for p in self.probe:
            p.append_sol(X[0])

        return True


class MG(IterativeSolver):   # LinearSolver
    def __init__(self, operators, smoothers, prolongations,
                 presmoother_count=1,
                 postsmoother_count=1):

        own_operators = [False]*len(operators)
        own_smoothers = [False]*len(smoothers)
        own_prolongations = [False]*len(prolongations)

        mg = mfem.Multigrid(operators, smoothers, prolongations,
                            own_operators, own_smoothers, own_prolongations)
        mg.SetCycleType(mfem.Multigrid.CycleType_VCYCLE,
                        presmoother_count, postsmoother_count)

        self.solver = mg
        #self.A = operators[-1]


def fill_prolongation_operator(engine, level, blk_opr):
    engine.access_idx = 0
    P = None
    diags = []

    A, _X, _RHS, _Ae,  _B,  _M, _dep_vars = engine.assembled_blocks
    use_complex_opr = A.complex and (A.shape[0] == blk_opr.NumRowBlocks())

    hights = A.get_local_row_heights()
    widths = A.get_local_col_widths()

    print(hights, widths)
    cols = [0]
    rows = [0]

    for dep_var in engine.r_dep_vars:
        offset = engine.r_dep_var_offset(dep_var)

        tmp_cols = []
        tmp_rows = []
        tmp_diags = []

        if use_complex_opr:
            mat = blk_opr._linked_op[(offset, offset)]
            conv = mat.GetConvention()
            conv == (1 if mfem.ComplexOperator.HERMITIAN else -1)
        else:
            conv = 1

        if engine.r_isFESvar(dep_var):
            h = engine.fespaces.get_hierarchy(dep_var)
            P = h.GetProlongationAtLevel(level)
            tmp_cols.append(P.Width())
            tmp_rows.append(P.Height())
            tmp_diags.append(P)
            if A.complex:
                tmp_cols.append(P.Width())
                tmp_rows.append(P.Height())
                if conv == -1:
                    oo2 = mfem.ScaleOperator(P, -1)
                    oo2._opr = P
                    tmp_diags.append(oo2)
                else:
                    tmp_diags.append(P)
        else:
            tmp_cols.append(widths[offset])
            tmp_rows.append(widths[offset])
            tmp_diags.append(mfem.IdentityOperator(widths[offset]))

            if A.complex:
                tmp_cols.append(widths[offset])
                tmp_rows.append(widths[offset])
                oo = mfem.IdentityOperator(widths[offset])
                if conv == -1:
                    oo2 = mfem.ScaleOperator(oo, -1)
                    oo2._opr = oo
                    tmp_diags.append(oo2)
                else:
                    tmp_diags.append(oo)
        if use_complex_opr:
            tmp_cols = [0] + tmp_cols
            tmp_rows = [0] + tmp_rows
            offset_c = mfem.intArray(tmp_cols)
            offset_r = mfem.intArray(tmp_rows)
            offset_c.PartialSum()
            offset_r.PartialSum()
            smoother = mfem.BlockOperator(offset_r, offset_c)
            smoother.SetDiagonalBlock(0, tmp_diags[0])
            smoother.SetDiagonalBlock(1, tmp_diags[1])
            smoother._smoother = tmp_diags
            cols.append(tmp_cols[1]*2)
            rows.append(tmp_rows[1]*2)
            diags.append(smoother)
        else:
            cols.extend(tmp_cols)
            rows.extend(tmp_rows)
            diags.extend(tmp_diags)

    ro = mfem.intArray(rows)
    co = mfem.intArray(cols)
    ro.PartialSum()
    co.PartialSum()

    P = mfem.BlockOperator(ro, co)
    for i, d in enumerate(diags):
        P.SetBlock(i, i, d)
    P._diags = diags
    return P


def genearate_smoother(engine, level, blk_opr):
    from petram.engine import ParallelEngine
    assert not isinstance(
        engine, ParallelEngine), "Parallel is not supported"

    engine.access_idx = 0
    P = None
    diags = []
    cols = [0]
    ess_tdofs = []
    A, _X, _RHS, _Ae,  _B,  _M, _dep_vars = engine.assembled_blocks
    widths = A.get_local_col_widths()

    use_complex_opr = A.complex and (A.shape[0] == blk_opr.NumRowBlocks())

    for dep_var in engine.r_dep_vars:
        offset = engine.r_dep_var_offset(dep_var)

        tmp_cols = []
        tmp_diags = []
        if engine.r_isFESvar(dep_var):
            ess_tdof = mfem.intArray(engine.ess_tdofs[dep_var])
            ess_tdofs.append(ess_tdofs)
            opr = A[offset, offset]

            if use_complex_opr:
                mat = blk_opr._linked_op[(offset, offset)]
                conv = mat.GetConvention()
                conv == 1 if mfem.ComplexOperator.HERMITIAN else -1
                mat1 = mat._real_operator
                mat2 = mat._imag_operator
            else:
                conv = 1
                if A.complex:
                    mat1 = blk_opr.GetBlock(offset*2, offset*2)
                    mat2 = blk_opr.GetBlock(offset*2 + 1, offset*2 + 1)
                else:
                    mat1 = blk_opr.GetBlock(offset, offset)

            if A.complex:
                dd = opr.diagonal()
                diag = mfem.Vector(list(dd.real))
                print("real", dd.real)
                rsmoother = mfem.OperatorChebyshevSmoother(mat1,
                                                           diag,
                                                           ess_tdof,
                                                           2)
                dd = opr.diagonal()*conv
                diag = mfem.Vector(list(dd.real))
                print("imag", dd.imag)
                ismoother = mfem.OperatorChebyshevSmoother(mat1,
                                                           diag,
                                                           ess_tdof,
                                                           2)
                tmp_cols.append(opr.shape[0])
                tmp_cols.append(opr.shape[0])
                tmp_diags.append(rsmoother)
                tmp_diags.append(ismoother)

            else:
                dd = opr.diagonal()
                diag = mfem.Vector(list(dd))
                smoother = mfem.OperatorChebyshevSmoother(mat1,
                                                          diag,
                                                          ess_tdof,
                                                          2)
                tmp_diags.append(smoother)
                tmp_cols.append(opr.shape[0])

        else:
            print("Non FESvar", dep_var, offset)
            tmp_cols.append(widths[offset])
            tmp_diags.append(mfem.IdentityOperator(widths[offset]))
            if A.complex:
                tmp_cols.append(widths[offset])
                oo = mfem.IdentityOperator(widths[offset])
                if conv == -1:
                    oo2 = mfem.ScaleOperator(oo, -1)
                    oo2._opr = oo
                    tmp_diags.append(oo2)
                else:
                    tmp_diags.append(oo)

        if use_complex_opr:
            tmp_cols = [0] + tmp_cols
            blockOffsets = mfem.intArray(tmp_cols)
            blockOffsets.PartialSum()
            smoother = mfem.BlockDiagonalPreconditioner(blockOffsets)
            smoother.SetDiagonalBlock(0, tmp_diags[0])
            smoother.SetDiagonalBlock(1, tmp_diags[1])
            smoother._smoother = tmp_diags
            cols.append(tmp_cols[1]*2)
            diags.append(smoother)
        else:
            cols.extend(tmp_cols)
            diags.extend(tmp_diags)

    co = mfem.intArray(cols)
    co.PartialSum()

    P = mfem.BlockDiagonalPreconditioner(co)
    for i, d in enumerate(diags):
        P.SetDiagonalBlock(i,  d)
    P._diags = diags
    P._ess_tdofs = ess_tdofs
    return P



import numpy as np

from petram.mfem_config import use_parallel

if use_parallel:
   import mfem.par as mfem
   from mpi4py import MPI
   num_proc = MPI.COMM_WORLD.size
   myid     = MPI.COMM_WORLD.rank
   from petram.helper.mpi_recipes import *   
else:
   import mfem.ser as mfem
   
import petram.debug
dprint1, dprint2, dprint3 = petram.debug.init_dprints('Mesh1Dl')

def straight_line_mesh(lengths, nsegs, filename='',
                       refine=False, fix_orientation=False,
                       sdim = 3, x0=0.0):

    Nvert = np.sum(nsegs)+1
    Nelem = np.sum(nsegs)
    Nbdrelem = len(lengths)+1
    mesh = mfem.Mesh(1, Nvert, Nelem,  Nbdrelem, sdim)

    ivert = {}
    L = np.hstack(([0], np.cumsum(lengths))).astype(float)
    P = np.hstack(([0], np.cumsum(nsegs))).astype(int)
    X = [np.linspace(L[i], L[i+1], n+1)[1:] for i, n in enumerate(nsegs)]
    X = np.hstack(([0], np.hstack(X)))
    A = np.hstack([[i+1]*n for i, n in enumerate(nsegs)])
    for k, i in enumerate(P):
        ptx = mfem.Point(i)
        ptx.SetAttribute(k+1)
        mesh.AddBdrElement(ptx)
        ptx.thisown = False
        
    for i in range(X.shape[0]-1):
        seg = mfem.Segment((i, i+1), A[i])
        mesh.AddElement(seg)
        seg.thisown = False        
    for i in range(X.shape[0]):
         pt = [0]*sdim
         pt[0] = X[i] + x0
         mesh.AddVertex(pt)
         
    mesh.FinalizeTopology()         
    mesh.Finalize(refine, fix_orientation)

    if filename != '':
        mesh.PrintToFile(filename, 8)
    return mesh

def quad_rectangle_mesh(xlengths, xnsegs, ylengths, ynsegs,
                        filename='', refine=False, fix_orientation=False,
                        sdim=3, x0=None):

    x0 = (0.0, 0.0) if x0 is None else x0
    
    Nvert = (np.sum(xnsegs)+1)*(np.sum(ynsegs)+1)
    Nelem = np.sum(xnsegs)*np.sum(ynsegs)
    Nbdrelem = (np.sum(xnsegs)*(len(ylengths)+1) + np.sum(ynsegs)*(len(xlengths)+1))
    mesh = mfem.Mesh(2, Nvert, Nelem,  Nbdrelem, sdim)
    
    Lx = np.hstack(([0], np.cumsum(xlengths)))
    x = [np.linspace(Lx[i], Lx[i+1], n+1)[:-1].astype(float) for i, n in enumerate(xnsegs)]+[np.sum(xlengths)]
    x = np.hstack(x)
    kx = np.hstack(([0], np.cumsum(xnsegs)))
    kkx = sum([[i]*n for i,n in enumerate(xnsegs)],[])
    
    Ly = np.hstack(([0], np.cumsum(ylengths)))
    y = [np.linspace(Ly[i], Ly[i+1], n+1)[:-1].astype(float) for i, n in enumerate(ynsegs)]+[np.sum(ylengths)]
    y = np.hstack(y)
    ky = np.hstack(([0], np.cumsum(ynsegs)))
    kky = sum([[i]*n for i,n in enumerate(ynsegs)],[])

    X, Y = np.meshgrid(x, y)

    X = X + x0[0]
    Y = Y + x0[1]
    
    IDX = np.arange(len(x)*len(y)).reshape(len(y), len(x))

    ax = np.hstack([[i]*n for i, n in enumerate(xnsegs)])
    ay = np.hstack([[j]*n for j, n in enumerate(ynsegs)])
    AX, AY = np.meshgrid(ax, ay)
    KDOM = AY*len(xnsegs) + AX + 1

    e_count = 1
    for i in range(len(y[:-1])):
        for j in range(len(x[:-1])):
            elem = [IDX[i, j], IDX[i, j+1], IDX[i+1, j+1], IDX[i+1, j]]
            mesh.AddQuad(elem, KDOM[i, j])
            e_count = e_count + 1

    kbdr = 1
    b_count = 1

    for k, j in enumerate(kx):
       for i in range(len(y[:-1])):
            elem = [IDX[i, j], IDX[i+1, j]]
            mesh.AddBdrSegment(elem, kbdr+kky[i])
            b_count = b_count + 1
       kbdr = kbdr + len(ynsegs)
        
    for k, i in enumerate(ky):
       for j in range(len(x[:-1])):
            elem = [IDX[i, j], IDX[i, j+1]]
            mesh.AddBdrSegment(elem, kbdr+kkx[j])
            b_count = b_count + 1            
       kbdr = kbdr + len(xnsegs)
        
    v_count = 1
    for xx, yy in zip(X.flatten(), Y.flatten()):
        ptx = [xx, yy, 0.0]
        mesh.AddVertex(ptx)
        v_count = v_count + 1
        
    #print(e_count, b_count, v_count)
    #print(Nelem,  Nbdrelem, Nvert)

    mesh.FinalizeTopology()         
    mesh.Finalize(refine, fix_orientation)

    if filename != '':
        mesh.PrintToFile(filename, 8)
    return mesh
 
def hex_box_mesh(xlengths, xnsegs, ylengths, ynsegs, zlengths, znsegs,
                 filename='', refine=False, fix_orientation=False,
                 sdim=3, x0=None):
   
    x0 = (0.0, 0.0, 0.0) if x0 is None else x0

    Nvert = (np.sum(xnsegs)+1)*(np.sum(ynsegs)+1)*(np.sum(znsegs)+1)
    Nelem = np.sum(xnsegs)*np.sum(ynsegs)*np.sum(znsegs)
    Nbdrelem = (np.sum(xnsegs)*np.sum(ynsegs)*(len(zlengths)+1) + 
                np.sum(xnsegs)*np.sum(znsegs)*(len(ylengths)+1) +
                np.sum(znsegs)*np.sum(ynsegs)*(len(xlengths)+1))

    mesh = mfem.Mesh(3, Nvert, Nelem,  Nbdrelem, sdim)
    
    Lx = np.hstack(([0], np.cumsum(xlengths)))
    x = [np.linspace(Lx[i], Lx[i+1], n+1)[:-1].astype(float) for i, n in enumerate(xnsegs)]+[np.sum(xlengths)]
    x = np.hstack(x)
    kx = np.hstack(([0], np.cumsum(xnsegs)))
    kkx = sum([[i]*n for i,n in enumerate(xnsegs)],[])
    
    Ly = np.hstack(([0], np.cumsum(ylengths)))
    y = [np.linspace(Ly[i], Ly[i+1], n+1)[:-1].astype(float) for i, n in enumerate(ynsegs)]+[np.sum(ylengths)]
    y = np.hstack(y)
    ky = np.hstack(([0], np.cumsum(ynsegs)))
    kky = sum([[i]*n for i,n in enumerate(ynsegs)],[])

    Lz = np.hstack(([0], np.cumsum(zlengths)))
    z = [np.linspace(Lz[i], Lz[i+1], n+1)[:-1].astype(float) for i, n in enumerate(znsegs)]+[np.sum(zlengths)]
    z = np.hstack(z)
    kz = np.hstack(([0], np.cumsum(znsegs)))
    kkz = sum([[i]*n for i,n in enumerate(znsegs)],[])
    
    X, Y, Z = np.meshgrid(x, y, z)   # y-idx, x-idx, z-idx

    X = X + x0[0]
    Y = Y + x0[1]
    Z = Z + x0[2]    
    
    IDX = np.arange(len(x)*len(y)*len(z)).reshape(len(y), len(x), len(z))

    ax = np.hstack([[i]*n for i, n in enumerate(xnsegs)])
    ay = np.hstack([[j]*n for j, n in enumerate(ynsegs)])
    az = np.hstack([[j]*n for j, n in enumerate(znsegs)])
    
    AX, AY, AZ = np.meshgrid(ax, ay, az)
    KDOM = AZ*len(xnsegs)*len(ynsegs) + AY*len(xnsegs) + AX + 1

    e_count = 1
    
    for i in range(len(y[:-1])):
        for j in range(len(x[:-1])):
            for k in range(len(z[:-1])):
               elem = [IDX[i, j, k ], IDX[i, j+1, k  ], IDX[i+1, j+1, k  ], IDX[i+1, j, k  ],
                       IDX[i, j,k+1], IDX[i, j+1, k+1], IDX[i+1, j+1, k+1], IDX[i+1, j, k+1]]
               mesh.AddHex(elem, KDOM[i, j, k])
               e_count = e_count + 1

    kbdr = 1
    b_count = 1

    for j in kx:
        for i in range(len(y[:-1])):
            for k in range(len(z[:-1])):
                elem = [IDX[i, j, k], IDX[i+1, j, k], IDX[i+1, j, k+1], IDX[i, j, k+1]] 
                mesh.AddBdrQuad(elem, kbdr+kky[i]*len(znsegs) + kkz[k])
                b_count = b_count + 1                
        kbdr = kbdr + len(ynsegs)*len(znsegs)


    for i in ky:
        for k in range(len(z[:-1])):                 
            for j in range(len(x[:-1])):
                elem = [IDX[i, j, k], IDX[i, j+1, k], IDX[i, j+1, k+1], IDX[i, j, k+1]]
                mesh.AddBdrQuad(elem, kbdr+kkz[k]*len(xnsegs) + kkx[j])
                b_count = b_count + 1                                                
        kbdr = kbdr + len(xnsegs)*len(znsegs)

    for k in kz:
        for j in range(len(x[:-1])):       
            for i in range(len(y[:-1])):
                elem = [IDX[i, j, k], IDX[i+1, j, k], IDX[i+1, j+1, k], IDX[i, j+1, k]]               
                mesh.AddBdrQuad(elem, kbdr+kkx[j]*len(ynsegs) + kky[i])
                b_count = b_count + 1                                
        kbdr = kbdr + len(xnsegs)*len(ynsegs)
       
    v_count = 1
    for xx, yy, zz in zip(X.flatten(), Y.flatten(), Z.flatten()):
        ptx = [xx, yy, zz]
        mesh.AddVertex(ptx)
        v_count = v_count + 1
    #print(e_count, b_count, v_count)
    #print(Nelem,  Nbdrelem, Nvert)

    mesh.FinalizeTopology()         
    mesh.Finalize(refine, fix_orientation)

    if filename != '':
        mesh.PrintToFile(filename, 8)
    return mesh





from __future__ import print_function
import petram.helper.pickle_wrapper as pickle

from itertools import product
import os
import numpy as np
import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('ParametricScanner')
dprint0 = debug.regular_print('ParametricScanner', True)
format_memory_usage = debug.format_memory_usage


class DefaultParametricScanner(object):
    def __init__(self, data=None):
        if data is None:
            data = []
        self.idx = 0
        self.target = None
        self.set_data(data)

    def set_data_from_model(self, model):
        '''
        this is called after __init__.
        model is passed. so that it can be set using   
        model tree
        '''
        pass

    def set_data(self, data):
        self._data = data
        self.max = len(data)

    def __iter__(self):
        self.idx = 0
        return self

    def __next__(self):

        if self.idx == self.max:
            raise StopIteration

        data = self._data[self.idx]
        dprint0("==== Entering next parameter:", data, "(" +
                str(self.idx+1) + "/" + str(self.max) + ")")
        dprint1(format_memory_usage())

        self.apply_param(data)

        self.idx = self.idx + 1
        return self.idx

    def list_data(self):
        return list(self._data)

    def set_phys_models(self, targets):
        '''
        set target physics model
        '''
        if (not isinstance(targets, tuple) and
                not isinstance(targets, list)):
            self.target_phys = [targets]
        else:
            self.target_phys = targets

    def next(self):
        return self.__next__()

    def __len__(self):
        return self.max

    def len(self):
        return self.max

    def save_scanner_data(self, solver):
        solver_name = solver.fullpath()
        data = self.list_data()
        dprint1("saving parameter", os.getcwd())
        try:
            from mpi4py import MPI
        except ImportError:
            from petram.helper.dummy_mpi import MPI
        myid = MPI.COMM_WORLD.rank

        if myid == 0:
            fid = open("parametric_data_"+solver_name, "wb")
            dd = {"name": solver_name, "data": data}
            pickle.dump(dd, fid)
            fid.close()

        MPI.COMM_WORLD.Barrier()

    def set_model(self, data):
        raise NotImplementedError(
            "set model for parametric scanner needs to be given in subclass")

    @property
    def names(self):
        '''
        suposed to return parameternames
        '''
        raise NotImplementedError(
            "set model for parametric scanner needs to be given in subclass")


class SimpleScanner(DefaultParametricScanner):
    '''
    Scan("freq", [3e9, 4e9, 5e9])
    # parameters are expanded to run all combination
    Scan("freq", [3e9, 4e9, 5e9], "phase", [0, 90])
    Scan("freq", [3e9, 4e9, 5e9], "phase", [0, 90, 180], product = False)
    # 1D scan
    Scan("freq", "phase", start = (3e9, 0), stop = (5e9, 180), nstep = 3)
    # 2D scan
    Scan("freq", "phase", start = (3e9, 0), stop = (5e9, 180), nstep = (3,4)) 
    '''

    def __init__(self, *args, **kwargs):
        use_product = kwargs.pop('product', True)

        if len(kwargs) != 0:
            self._names = args
            starts = np.atleast_1d(kwargs['start'])
            stops = np.atleast_1d(kwargs['stop'])
            steps = np.atleast_1d(kwargs['nstep'])
            if len(steps) == 1:
                use_product = False

            data = []
            for k in range(len(self._names)):
                s = starts[k]
                e = stops[k]
                n = steps[0] if len(steps) == 1 else steps[k]
                data.append(np.linspace(s, e, n))

        else:
            o = iter(args)
            names = []
            data = []
            while True:
                try:
                    names.append(o.__next__())
                    data.append(o.__next__())
                except StopIteration:
                    break
            self._names = names

        if use_product:
            data = product(*data)
        else:
            data = zip(*data)

        DefaultParametricScanner.__init__(self, data=list(data))

    def apply_param(self, data):
        names = self._names

        dprint1("Simple Scanner: Target " + str(self.target_phys))
        for k, name in enumerate(names):
            id_list = []
            for phys in self.target_phys:
                if id(phys._global_ns) in id_list:
                    continue

                dprint1("Simple Scanner: Setting " + name + ':' + str(data[k]) +
                        " setting in " + str(phys))
                phys._global_ns[name] = data[k]
                id_list.append(id(phys._global_ns))

    @property
    def names(self):
        '''
        suposed to return parameternames
        '''
        return self._names


Scan = SimpleScanner
